"""Methods regarding feature extractors."""
from __future__ import annotations
from pathlib import Path
import ast
import subprocess
from sparkle.types import SparkleCallable, SolverStatus
from sparkle.structures import FeatureDataFrame
from sparkle.tools.runsolver_parsing import get_status


class Extractor(SparkleCallable):
    """Extractor base class for extracting features from instances."""
    wrapper = "sparkle_extractor_wrapper.py"

    def __init__(self: Extractor,
                 directory: Path,
                 runsolver_exec: Path = None,
                 raw_output_directory: Path = None,
                 ) -> None:
        """Initialize solver.

        Args:
            directory: Directory of the solver.
            runsolver_exec: Path to the runsolver executable.
                By default, runsolver in directory.
            raw_output_directory: Directory where solver will write its raw output.
                Defaults to directory / tmp
        """
        super().__init__(directory, runsolver_exec, raw_output_directory)
        self._features = None
        self._feature_groups = None
        self._output_dimension = None
        self._groupwise_computation = None

    @property
    def features(self: Extractor) -> list[tuple[str, str]]:
        """Determines the features of the extractor."""
        if self._features is None:
            extractor_process = subprocess.run(
                [self.directory / Extractor.wrapper, "-features"], capture_output=True)
            self._features = ast.literal_eval(extractor_process.stdout.decode())
        return self._features

    @property
    def feature_groups(self: Extractor) -> list[str]:
        """Returns the various feature groups the Extractor has."""
        if self._feature_groups is None:
            self._feature_groups = list(set([group for group, _ in self.features]))
        return self._feature_groups

    @property
    def output_dimension(self: Extractor) -> int:
        """The size of the output vector of the extractor."""
        return len(self.features)

    @property
    def groupwise_computation(self: Extractor) -> bool:
        """Determines if you can call the extractor per group for parallelisation."""
        if self._groupwise_computation is None:
            extractor_help = subprocess.run([self.directory / Extractor.wrapper, "-h"],
                                            capture_output=True)
            # Not the cleanest / most precise way to determine this
            self._groupwise_computation =\
                "-feature_group" in extractor_help.stdout.decode()
        return self._groupwise_computation

    def build_cmd(self: Extractor,
                  instance: Path | list[Path],
                  feature_group: str = None,
                  output_file: Path = None,
                  runsolver_args: list[str | Path] = None,
                  ) -> list[str]:
        """Builds a command line string seperated by space.

        Args:
            instance: The instance to run on
            feature_group: The optional feature group to run the extractor for.
            outputfile: Optional file to write the output to.
            runsolver_args: The arguments for runsolver. If not present,
                will run the extractor without runsolver.

        Returns:
            The command seperated per item in the list.
        """
        cmd_list_extractor = []
        if not isinstance(instance, list):
            instance = [instance]
        if runsolver_args is not None:
            # Ensure stringification of runsolver configuration is done correctly
            cmd_list_extractor += [str(self.runsolver_exec.absolute())]
            cmd_list_extractor += [str(runsolver_config) for runsolver_config
                                   in runsolver_args]
        cmd_list_extractor += [f"{self.directory / Extractor.wrapper}",
                               "-extractor_dir", f"{self.directory}/",
                               "-instance_file"] + [str(file) for file in instance]
        if feature_group is not None:
            cmd_list_extractor += ["-feature_group", feature_group]
        if output_file is not None:
            cmd_list_extractor += ["-output_file", str(output_file)]
        return cmd_list_extractor

    def run(self: Extractor,
            instance: Path | list[Path],
            feature_group: str = None,
            output_file: Path = None,
            runsolver_args: list[str | Path] = None) -> list | None:
        """Runs an extractor job with Runrunner.

        Args:
            extractor_path: Path to the executable
            instance: Path to the instance to run on
            feature_group: The feature group to compute. Must be supported by the
                extractor to use.
            output_file: Target output. If None, piped to the RunRunner job.
            runsolver_args: List of run solver args, each word a seperate item.

        Returns:
            The features or None if an output file is used, or features can not be found.
        """
        if feature_group is not None and not self.groupwise_computation:
            # This extractor cannot handle groups, compute all features
            feature_group = None
        cmd_extractor = self.build_cmd(
            instance, feature_group, output_file, runsolver_args)
        extractor = subprocess.run(cmd_extractor, capture_output=True)
        if output_file is None:
            try:
                features = ast.literal_eval(extractor.stdout.decode())
                return features
            except Exception:
                return None
        return None

    def get_feature_vector(self: Extractor,
                           result: Path,
                           runsolver_values: Path = None) -> list[str]:
        """Extracts feature vector from an output file.

        Args:
            result: The raw output of the extractor
            runsolver_values: The output of runsolver.

        Returns:
            A list of features. Vector of missing values upon failure.
        """
        if result.exists() and get_status(runsolver_values,
                                          None) != SolverStatus.TIMEOUT:
            feature_values = ast.literal_eval(result.read_text())
            return [str(value) for _, _, value in feature_values]
        return [FeatureDataFrame.missing_value] * self.output_dimension



#!/usr/bin/env python3
"""Sparkle command for the computation of the marginal contributions."""
import sys
import argparse
from pathlib import Path
import operator

import tabulate

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.types import SparkleObjective


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.PerfectSelectorMarginalContributionArgument.names,
                        **ac.PerfectSelectorMarginalContributionArgument.kwargs)
    parser.add_argument(*ac.ActualMarginalContributionArgument.names,
                        **ac.ActualMarginalContributionArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)

    return parser


def compute_selector_performance(
        actual_portfolio_selector: Path,
        performance_data: PerformanceDataFrame,
        feature_data: FeatureDataFrame,
        objective: SparkleObjective) -> float:
    """Return the performance of a selector over all instances.

    Args:
      actual_portfolio_selector: Path to portfolio selector.
      performance_data: The performance data.
      feature_data: The feature data.
      objective: Objective to compute the performance for

    Returns:
      The selector performance as a single floating point number.
    """
    performance_path = actual_portfolio_selector.parent / "performance.csv"
    if performance_path.exists():
        selector_performance_data = PerformanceDataFrame(performance_path)
        return objective.instance_aggregator(
            selector_performance_data.get_values("portfolio_selector"))
    selector_performance_data = performance_data.copy()

    selector_performance_data.add_solver("portfolio_selector")
    selector_performance_data.csv_filepath =\
        actual_portfolio_selector.parent / "performance.csv"
    selector = gv.settings().get_general_sparkle_selector()

    schedule = {}
    for instance in performance_data.instances:
        # We get the performance for an instance by infering the model predicition
        # for the instance.
        feature_vector = feature_data.get_instance(instance)
        schedule[instance] = selector.run(actual_portfolio_selector, feature_vector)
    schedule_performance = selector_performance_data.schedule_performance(
        schedule, target_solver="portfolio_selector", objective=objective)
    # Remove solvers from the dataframe
    selector_performance_data.remove_solver(performance_data.solvers)
    selector_performance_data.save_csv()  # Save the results to disk
    return objective.instance_aggregator(schedule_performance)


def compute_selector_marginal_contribution(
        performance_data: PerformanceDataFrame,
        feature_data: FeatureDataFrame,
        selector_scenario: Path,
        objective: SparkleObjective) -> list[tuple[str, float]]:
    """Compute the marginal contributions of solvers in the selector.

    Args:
      performance_data: Performance data object
      feature_data_csv_path: Path to the CSV file with the feature data.
      selector_scenario: Path to the selector scenario for which to compute
        marginal contribution.
      objective: Objective to compute the marginal contribution for.

    Returns:
      A list of 2-tuples where every 2-tuple is of the form
        (solver name, marginal contribution, best_performance).
    """
    portfolio_selector_path = selector_scenario / "portfolio_selector"

    if not portfolio_selector_path.exists():
        print(f"ERROR: Selector {portfolio_selector_path} does not exist! "
              "Cannot compute marginal contribution.")
        sys.exit(-1)

    selector_performance = compute_selector_performance(
        portfolio_selector_path, performance_data,
        feature_data, objective)

    rank_list = []
    compare = operator.lt if objective.minimise else operator.gt
    # Compute contribution per solver
    # NOTE: This could be parallelised
    for solver in performance_data.solvers:
        solver_name = Path(solver).name
        # 1. Copy the dataframe original df
        tmp_performance_df = performance_data.copy()
        # 2. Remove the solver from this copy
        tmp_performance_df.remove_solver(solver)
        ablated_actual_portfolio_selector =\
            selector_scenario / f"ablate_{solver_name}" / "portfolio_selector"
        if not ablated_actual_portfolio_selector.exists():
            print(f"WARNING: Selector without {solver_name} does not exist! "
                  f"Cannot compute marginal contribution of {solver_name}.")
            continue

        ablated_selector_performance = compute_selector_performance(
            ablated_actual_portfolio_selector, tmp_performance_df,
            feature_data, objective)

        # 1. If the performance remains equal, this solver did not contribute
        # 2. If there is a performance decay without this solver, it does contribute
        # 3. If there is a performance improvement, we have a bad portfolio selector
        if ablated_selector_performance == selector_performance:
            marginal_contribution = 0.0
        elif not compare(ablated_selector_performance, selector_performance):
            # In the case that the performance decreases, we have a contributing solver
            marginal_contribution = ablated_selector_performance / selector_performance
        else:
            print("****** WARNING DUBIOUS SELECTOR/SOLVER: "
                  f"The omission of solver {solver_name} yields an improvement. "
                  "The selector improves better without this solver. It may be usefull "
                  "to construct a portfolio without this solver.")
            marginal_contribution = 0.0

        rank_list.append((solver, marginal_contribution, ablated_selector_performance))

    rank_list.sort(key=lambda contribution: contribution[1], reverse=True)
    return rank_list


def compute_marginal_contribution(
        scenario: Path,
        compute_perfect: bool, compute_actual: bool) -> None:
    """Compute the marginal contribution.

    Args:
        scenario: Path to the selector scenario for which to compute
        compute_perfect: Bool indicating if the contribution for the perfect
            portfolio selector should be computed.
        compute_actual: Bool indicating if the contribution for the actual portfolio
             selector should be computed.
    """
    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
    objective = gv.settings().get_general_sparkle_objectives()[0]

    if compute_perfect:
        # Perfect selector is the computation of the best performance per instance
        print("Computing each solver's marginal contribution to perfect selector ...")
        contribution_data = performance_data.marginal_contribution(
            objective=objective.name, sort=True)
        table = tabulate.tabulate(
            contribution_data,
            headers=["Solver", "Marginal Contribution", "Best Performance"],)
        print(table, "\n")
        print("Marginal contribution (perfect selector) computing done!")

    if compute_actual:
        print("Start computing marginal contribution per Solver to actual selector...")
        contribution_data = compute_selector_marginal_contribution(
            performance_data,
            feature_data,
            scenario,
            objective
        )
        table = tabulate.tabulate(
            contribution_data,
            headers=["Solver", "Marginal Contribution", "Best Performance"],)
        print(table, "\n")
        print("Marginal contribution (actual selector) computing done!")


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.COMPUTE_MARGINAL_CONTRIBUTION]
    )

    if ac.set_by_user(args, "settings_file"):
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )  # Do first, so other command line options can override settings from the file
    if ac.set_by_user(args, "objectives"):
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE
        )
    selection_scenario = gv.latest_scenario().get_selection_scenario_path()

    if not (args.perfect | args.actual):
        print("ERROR: compute_marginal_contribution called without a flag set to"
              " True, stopping execution")
        sys.exit(-1)

    compute_marginal_contribution(selection_scenario, args.perfect, args.actual)

    # Write used settings to file
    gv.settings().write_used_settings()



#!/usr/bin/env python3
"""Sparkle class to organise configuration output."""

from __future__ import annotations

from sparkle.platform import generate_report_for_parallel_portfolio as sgrfpp
from sparkle.instance import InstanceSet
from sparkle.platform.output.structures import ParallelPortfolioResults
from sparkle.types import SparkleObjective

import json
from pathlib import Path
import csv


class ParallelPortfolioOutput:
    """Class that collects parallel portfolio data and outputs it a JSON format."""

    def __init__(self: ParallelPortfolioOutput, parallel_portfolio_path: Path,
                 instance_set: InstanceSet,
                 objective: SparkleObjective,
                 output: Path) -> None:
        """Initialize ParallelPortfolioOutput class.

        Args:
            parallel_portfolio_path: Path to parallel portfolio output directory
            instance_set: List of instances
            objective: The objective of the portfolio
            output: Path to the output directory
        """
        if not output.is_file():
            self.output = output / "parallel_portfolio.json"
        else:
            self.output = output

        self.instance_set = instance_set
        csv_data = [line for line in
                    csv.reader((parallel_portfolio_path / "results.csv").open("r"))]
        header = csv_data[0]
        csv_data = csv_data[1:]
        solver_column = header.index("Solver")
        instance_column = header.index("Instance")
        status_column = header.index("status")
        objective_column = header.index(objective.name)
        self.solver_list = list(set([line[solver_column] for line in csv_data]))

        # Collect solver performance for each instance
        instance_results = {name: [] for name in instance_set._instance_names}
        for row in csv_data:
            if row[instance_column] in instance_results.keys():
                instance_results[row[instance_column]].append(
                    [row[solver_column], row[status_column], row[objective_column]])

        solvers_solutions = self.get_solver_solutions(self.solver_list, csv_data)
        unsolved_instances = self.instance_set.size - sum([solvers_solutions[key]
                                                           for key in solvers_solutions])
        # sbs_runtime is redundant, the same information is available in instance_results
        _, sbs, runtime_all_solvers, _ =\
            sgrfpp.get_portfolio_metrics(self.solver_list,
                                         instance_set,
                                         instance_results,
                                         objective)

        self.results = ParallelPortfolioResults(unsolved_instances,
                                                sbs, runtime_all_solvers,
                                                instance_results)

    def get_solver_solutions(self: ParallelPortfolioOutput,
                             solver_list: list[str],
                             csv_data: list[list[str]]) -> dict:
        """Return dictionary with solution count for each solver."""
        # Default initalisation, increase solution counter for each successful evaluation
        solvers_solutions = {solver: 0 for solver in solver_list}
        instance_names_copy = self.instance_set._instance_names.copy()

        for line in csv_data:
            if line[0] in instance_names_copy and line[2].lower() == "success":
                solvers_solutions[line[1]] += 1
                instance_names_copy.remove(line[0])

        return solvers_solutions

    def serialize_instances(self: ParallelPortfolioOutput,
                            instances: list[InstanceSet]) -> dict:
        """Transform Instances to dictionary format."""
        # Even though parallel portfolio currently doesn't support multi sets,
        # this function is already ready to support mutliple sets
        return {
            "number_of_instance_sets": len(instances),
            "instance_sets": [
                {
                    "name": instance.name,
                    "number_of_instances": instance.size
                }
                for instance in instances
            ]
        }

    def serialize_results(self: ParallelPortfolioOutput,
                          pr: ParallelPortfolioResults) -> dict:
        """Transform results to dictionary format."""
        return {
            "sbs": pr.sbs,
            "unsolved_instances": pr.unsolved_instances,
            "runtime_solvers": pr.runtime_solvers,
            "solvers_performance": pr.solver_performance,
            "instance_results": pr.instance_results,
        }

    def write_output(self: ParallelPortfolioOutput) -> None:
        """Write data into a JSON file."""
        output_data = {
            "number_of_solvers": len(self.solver_list),
            "solvers": self.solver_list,
            "instances": self.serialize_instances([self.instance_set]),
            "results": self.serialize_results(self.results),
        }

        self.output.parent.mkdir(parents=True, exist_ok=True)
        with self.output.open("w") as f:
            json.dump(output_data, f, indent=4)



import numpy as np

from autofolio.facade.af_csv_facade import AFCsvFacade

__author__ = "Marius Lindauer"
__license__ = "BSD"
__version__ = "2.1.0"

perf_fn = "perf.csv"
feat_fn = "feats.csv"

# will be created (or overwritten) by AutoFolio
model_fn = "af_model.pkl"

af = AFCsvFacade(perf_fn=perf_fn, feat_fn=feat_fn)

# fit AutoFolio; will use default hyperparameters of AutoFolio
af.fit()

# tune AutoFolio's hyperparameter configuration for 4 seconds
config = af.tune(wallclock_limit=4)

# evaluate configuration using a 10-fold cross validation
score = af.cross_validation(config=config)

# re-fit AutoFolio using the (hopefully) better configuration
# and save model to disk
af.fit(config=config, save_fn=model_fn)

# load AutoFolio model and
# get predictions for new meta-feature vector
pred = AFCsvFacade.load_and_predict(vec=np.array([1.]), load_fn=model_fn)

print(pred)





#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Sparkle command to execute a parallel algorithm portfolio."""

import sys
import argparse
import random
import time
import shutil
import csv
import itertools
from pathlib import Path, PurePath

from tqdm import tqdm

import runrunner as rrr
from runrunner.base import Runner
from runrunner.slurm import Status

from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import global_variables as gv
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.solver import Solver
from sparkle.instance import instance_set, InstanceSet
from sparkle.types import SolverStatus, resolve_objective, UseTime


def run_parallel_portfolio(instances_set: InstanceSet,
                           portfolio_path: Path,
                           solvers: list[Solver],
                           run_on: Runner = Runner.SLURM) -> None:
    """Run the parallel algorithm portfolio.

    Args:
        instances_set: Set of instances to run on.
        portfolio_path: Path to the parallel portfolio.
        solvers: List of solvers to run on the instances.
        run_on: Currently only supports Slurm.
    """
    num_solvers, num_instances = len(solvers), len(instances_set._instance_paths)
    seeds_per_solver = gv.settings().get_parallel_portfolio_number_of_seeds_per_solver()
    num_jobs = num_solvers * num_instances * seeds_per_solver
    parallel_jobs = min(gv.settings().get_number_of_jobs_in_parallel(), num_jobs)
    if parallel_jobs > num_jobs:
        print("WARNING: Not all jobs will be started at the same time due to the "
              "limitation of number of Slurm jobs that can be run in parallel. Check"
              " your Sparkle Slurm Settings.")
    print(f"Sparkle parallel portfolio is running {seeds_per_solver} seed(s) per solver "
          f"on {num_solvers} solvers for {num_instances} instances ...")
    cmd_list = []
    cutoff = gv.settings().get_general_target_cutoff_time()
    objectives = gv.settings().get_general_sparkle_objectives()
    # Create a command for each instance-solver-seed combination
    for instance, solver in itertools.product(instances_set._instance_paths, solvers):
        for _ in range(seeds_per_solver):
            seed = int(random.getrandbits(32))
            solver_call_list = solver.build_cmd(
                instance.absolute(),
                objectives=objectives,
                seed=seed,
                cutoff_time=cutoff)
            cmd_list.append((" ".join(solver_call_list)).replace("'", '"'))
    # Jobs are added in to the runrunner object in the same order they are provided
    sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
    run = rrr.add_to_queue(
        runner=run_on,
        cmd=cmd_list,
        name=CommandName.RUN_PARALLEL_PORTFOLIO,
        parallel_jobs=parallel_jobs,
        path=portfolio_path,
        base_dir=sl.caller_log_dir,
        srun_options=["-N1", "-n1"] + sbatch_options,
        sbatch_options=sbatch_options
    )
    check_interval = gv.settings().get_parallel_portfolio_check_interval()
    instances_done = [False] * num_instances
    # We record the 'best' of all seed results per solver-instance,
    # setting start values for objectives that are always present
    default_objective_values = {}
    for o in objectives:
        default_value = float(sys.maxsize) if o.minimise else 0
        # Default values for time objectives can be linked to cutoff time
        if o.time:
            default_value = cutoff + 1.0
            if o.post_process is not None:
                default_value = o.post_process(default_value, cutoff)
        default_objective_values[o.name] = default_value
    job_output_dict = {instance_name: {solver.name: default_objective_values.copy()
                                       for solver in solvers}
                       for instance_name in instances_set._instance_names}
    n_instance_jobs = num_solvers * seeds_per_solver

    with tqdm(total=len(instances_done)) as pbar:
        pbar.set_description("Instances done")
        while not all(instances_done):
            prev_done = sum(instances_done)
            time.sleep(check_interval)
            job_status_list = [r.status for r in run.jobs]
            job_status_completed = [status == Status.COMPLETED
                                    for status in job_status_list]
            # The jobs are sorted by instance
            for i, instance in enumerate(instances_set._instance_paths):
                if instances_done[i]:
                    continue
                instance_job_slice = slice(i * n_instance_jobs,
                                           (i + 1) * n_instance_jobs)
                if any(job_status_completed[instance_job_slice]):
                    instances_done[i] = True
                    # Kill all running jobs for this instance
                    solver_kills = [0] * num_solvers
                    for job_index in range(i * n_instance_jobs,
                                           (i + 1) * n_instance_jobs):
                        if not job_status_completed[job_index]:
                            run.jobs[job_index].kill()
                            solver_index = int(
                                (job_index % n_instance_jobs) / seeds_per_solver)
                            solver_kills[solver_index] += 1
                    for solver_index in range(num_solvers):
                        # All seeds of a solver were killed on instance, set status kill
                        if solver_kills[solver_index] == seeds_per_solver:
                            solver_name = solvers[solver_index].name
                            job_output_dict[instance.name][solver_name]["status"] =\
                                SolverStatus.KILLED
            pbar.update(sum(instances_done) - prev_done)

    # Attempt to verify that all logs have been written (Slurm I/O latency)
    for index, cmd in enumerate(cmd_list):
        runsolver_configuration = cmd.split(" ")[:11]
        logs = [portfolio_path / p for p in runsolver_configuration
                if Path(p).suffix in [".log", ".val", ".rawres"]]
        if not all([p.exists() for p in logs]):
            time.sleep(check_interval)

    # Now iterate over runsolver logs to get runtime, get the lowest value per seed
    for index, cmd in enumerate(cmd_list):
        runsolver_configuration = cmd.split(" ")[:11]
        solver_output = Solver.parse_solver_output(run.jobs[i].stdout,
                                                   runsolver_configuration,
                                                   portfolio_path)
        solver_index = int((index % n_instance_jobs) / seeds_per_solver)
        solver_name = solvers[solver_index].name
        instance_name = instances_set._instance_names[int(index / n_instance_jobs)]
        cpu_time = solver_output["cpu_time"]
        cmd_output = job_output_dict[instance_name][solver_name]
        if cpu_time > 0.0 and cpu_time < cmd_output["cpu_time"]:
            for key, value in solver_output.items():
                if key in [o.name for o in objectives]:
                    job_output_dict[instance_name][solver_name][key] = value
            if "status" not in cmd_output or cmd_output["status"] != SolverStatus.KILLED:
                cmd_output["status"] = solver_output["status"]

    # Fix the CPU/WC time for non existent logs to instance min time + check_interval
    for instance in job_output_dict.keys():
        no_log_solvers = []
        min_time = cutoff
        for solver in job_output_dict[instance].keys():
            cpu_time = job_output_dict[instance][solver]["cpu_time"]
            if cpu_time == -1.0 or cpu_time == float(sys.maxsize):
                no_log_solvers.append(solver)
            elif cpu_time < min_time:
                min_time = cpu_time
        for solver in no_log_solvers:
            job_output_dict[instance][solver]["cpu_time"] = min_time + check_interval
            job_output_dict[instance][solver]["wall_time"] = min_time + check_interval
            # Fix runtime objectives with resolved CPU/Wall times
            for key, value in job_output_dict[instance][solver].items():
                objective = resolve_objective(key)
                if objective is not None and objective.time:
                    if objective.use_time == UseTime.CPU_TIME:
                        value = job_output_dict[instance][solver]["cpu_time"]
                    else:
                        value = job_output_dict[instance][solver]["wall_time"]
                    if objective.post_process is not None:
                        value = objective.post_process(value, cutoff)
                    job_output_dict[instance][solver][key] = value

    for index, instance_name in enumerate(instances_set._instance_names):
        index_str = f"[{index + 1}/{num_instances}] "
        instance_output = job_output_dict[instance_name]
        if all([instance_output[k]["status"] == SolverStatus.TIMEOUT
                for k in instance_output.keys()]):
            print(f"\n{index_str}{instance_name} was not solved within the cutoff-time.")
            continue
        print(f"\n{index_str}{instance_name} yielded the following Solver results:")
        for sindex in range(index * num_solvers, (index + 1) * num_solvers):
            solver_name = solvers[sindex % num_solvers].name
            job_info = job_output_dict[instance_name][solver_name]
            print(f"\t- {solver_name} ended with status {job_info['status']} in "
                  f"{job_info['cpu_time']}s CPU-Time ({job_info['wall_time']}s WC-Time)")

    # Write the results to a CSV
    csv_path = portfolio_path / "results.csv"
    values_header = ["status"] + [o.name for o in objectives]
    header = ["Instance", "Solver"] + values_header
    result_rows = [header]
    for instance_name in job_output_dict.keys():
        for solver_name in job_output_dict[instance_name].keys():
            job_o = job_output_dict[instance_name][solver_name]
            values = [instance_name, solver_name] + [
                job_o[key] if key in job_o else "None"
                for key in values_header]
            result_rows.append(values)
    with csv_path.open("w") as out:
        writer = csv.writer(out)
        writer.writerows(result_rows)


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
        parser: The parser with the parsed command line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.InstancePath.names,
                        **ac.InstancePath.kwargs)
    parser.add_argument(*ac.NicknamePortfolioArgument.names,
                        **ac.NicknamePortfolioArgument.kwargs)
    parser.add_argument(*ac.SolversArgument.names,
                        **ac.SolversArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.CutOffTimeArgument.names,
                        **ac.CutOffTimeArgument.kwargs)
    parser.add_argument(*ac.SolverSeedsArgument.names,
                        **ac.SolverSeedsArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    if args.solvers is not None:
        solver_paths = [resolve_object_name("".join(s),
                                            target_dir=gv.settings().DEFAULT_solver_dir)
                        for s in args.solvers]
        if None in solver_paths:
            print("Some solvers not recognised! Check solver names:")
            for i, name in enumerate(solver_paths):
                if solver_paths[i] is None:
                    print(f'\t- "{solver_paths[i]}" ')
            sys.exit(-1)
        solvers = [Solver(p) for p in solver_paths]
    else:
        solvers = [Solver(p) for p in
                   gv.settings().DEFAULT_solver_dir.iterdir() if p.is_dir()]

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.RUN_PARALLEL_PORTFOLIO])

    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    # Do first, so other command line options can override settings from the file
    if args.settings_file is not None:
        gv.settings().read_settings_ini(args.settings_file, SettingState.CMD_LINE)

    portfolio_path = args.portfolio_name

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    if args.solver_seeds is not None:
        gv.settings().set_parallel_portfolio_number_of_seeds_per_solver(
            args.solver_seeds, SettingState.CMD_LINE)

    if run_on == Runner.LOCAL:
        print("Parallel Portfolio is not fully supported yet for Local runs. Exiting.")
        sys.exit(-1)

    # Retrieve instance set
    data_set = resolve_object_name(
        args.instance_path,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir,
        instance_set)
    print(f"Running on {data_set.size} instance(s)...")

    if args.cutoff_time is not None:
        gv.settings().set_general_target_cutoff_time(args.cutoff_time,
                                                     SettingState.CMD_LINE)

    if args.objectives is not None:
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE)
    if not gv.settings().get_general_sparkle_objectives()[0].time:
        print("ERROR: Parallel Portfolio is currently only relevant for "
              "RunTime objectives. In all other cases, use validation")
        sys.exit(-1)

    if args.portfolio_name is not None:  # Use a nickname
        portfolio_path = gv.settings().DEFAULT_parallel_portfolio_output_raw /\
            args.portfolio_name
    else:  # Generate a timestamped nickname
        timestamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.gmtime(time.time()))
        randintstamp = int(random.getrandbits(32))
        portfolio_path = gv.settings().DEFAULT_parallel_portfolio_output_raw /\
            f"{timestamp}_{randintstamp}"
    if portfolio_path.exists():
        print(f"[WARNING] Portfolio path {portfolio_path} already exists! "
              "Overwrite? [y/n] ", end="")
        user_input = input()
        if user_input != "y":
            sys.exit()
        shutil.rmtree(portfolio_path)
    portfolio_path.mkdir(parents=True)
    run_parallel_portfolio(data_set, portfolio_path, solvers, run_on=run_on)

    # Update latest scenario
    gv.latest_scenario().set_parallel_portfolio_path(portfolio_path)
    gv.latest_scenario().set_latest_scenario(Scenario.PARALLEL_PORTFOLIO)
    gv.latest_scenario().set_parallel_portfolio_instance_path(args.instance_path)
    # Write used scenario to file
    gv.latest_scenario().write_scenario_ini()
    # Write used settings to file
    gv.settings().write_used_settings()
    print("Running Sparkle parallel portfolio is done!")



"""Test public methods of sparkle types."""

from __future__ import annotations
import unittest
import itertools

from sparkle.types import resolve_objective
from sparkle.types.objective import SparkleObjective, PAR


class TestResolveObjective(unittest.TestCase):
    """Test resolve_objective function."""

    def setUp(self: TestResolveObjective) -> None:
        """List for good and bad objective names."""
        self.test_names_good = ["quality", "quality10", "f1", "F1", "PAR10special",
                                "very_special-indicator"]
        self.test_names_bad = ["", "10", ".test", ":min", "quality:minimum",
                               "quality:min:max", "_test", "-accuracy"]

    def test_resolving_basic(self: TestResolveObjective) -> None:
        """Test successful strings that should result in a normal SparkleObjective."""
        for name in self.test_names_good:
            objective = resolve_objective(name)
            self.assertIsInstance(objective, SparkleObjective)
            self.assertEqual(objective.name, name)
            self.assertTrue(objective.minimise)

    def test_min_max_suffix(self: TestResolveObjective) -> None:
        """Test the min/max suffix."""
        for name, minimise in itertools.product(self.test_names_good, [True, False]):
            input_name = name
            input_name += ":min" if minimise else ":max"
            objective = resolve_objective(input_name)
            self.assertIsInstance(objective, SparkleObjective)
            self.assertEqual(objective.name, input_name)
            self.assertEqual(objective.minimise, minimise)

    def test_failing_strings(self: TestResolveObjective) -> None:
        """Test object names that should fail and return None."""
        for name in self.test_names_bad:
            self.assertIsNone(resolve_objective(name))

    def test_par(self: TestResolveObjective) -> None:
        """Test PAR objects that should successfully initialize."""
        for name, k, oname in [("PAR", 10, "PAR10"),
                               ("PAR10", 10, "PAR10"),
                               ("PAR2", 2, "PAR2"),
                               ("PAR100", 100, "PAR100"),
                               ("PAR10:max", 10, "PAR10"),
                               ("PAR10:min", 10, "PAR10"), ]:
            objective = resolve_objective(name)
            self.assertIsInstance(objective, SparkleObjective)
            self.assertIsInstance(objective, PAR)
            self.assertEqual(objective.name, oname)
            self.assertEqual(objective.k, k)
            self.assertTrue(objective.minimise)

    def test_failing_par(self: TestResolveObjective) -> None:
        """Test PAR objects names with k<0."""
        for name in ["PAR0", "PAR-1"]:
            print(f"{name=}")
            with self.assertRaises(ValueError):
                resolve_objective(name)



#!/usr/bin/env python3
"""Sparkle class to organise configuration output."""

from __future__ import annotations

from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.platform import generate_report_for_selection as sgfs
from sparkle.types.objective import SparkleObjective
from sparkle.instance import InstanceSet
from sparkle.platform.output.structures import SelectionPerformance, SelectionSolverData

import json
from pathlib import Path


class SelectionOutput:
    """Class that collects selection data and outputs it a JSON format."""

    def __init__(self: SelectionOutput, selection_scenario: Path,
                 train_data: PerformanceDataFrame,
                 feature_data: FeatureDataFrame,
                 training_instances: list[InstanceSet],
                 test_instances: list[InstanceSet],
                 objective: SparkleObjective,
                 cutoff_time: int,
                 output: Path) -> None:
        """Initialize SelectionOutput class.

        Args:
            selection_scenario: Path to selection output directory
            train_data: The performance input data for the selector
            feature_data: Feature data created by extractor
            training_instances: The set of training instances
            test_instances: The set of test instances
            objective: The objective of the selector
            cutoff_time: The cutoff time
            penalised_time: The penalised time
            output: Path to the output directory
        """
        if not output.is_file():
            self.output = output / "selection.json"
        else:
            self.output = output
        if test_instances is not None and not isinstance(test_instances, list):
            test_instances = [test_instances]

        self.training_instances = training_instances
        self.test_instances = test_instances
        self.cutoff_time = cutoff_time

        self.objective = objective
        self.solver_data = self.get_solver_data(train_data, self.objective)
        # Collect marginal contribution data
        self.marginal_contribution_perfect = train_data.marginal_contribution(objective,
                                                                              sort=True)
        self.marginal_contribution_actual = \
            sgfs.compute_selector_marginal_contribution(train_data,
                                                        feature_data,
                                                        selection_scenario,
                                                        objective)

        # Collect performance data
        portfolio_selector_performance_path = selection_scenario / "performance.csv"
        vbs_performance = objective.instance_aggregator(
            train_data.best_instance_performance(objective=objective.name))
        self.performance_data = SelectionPerformance(
            portfolio_selector_performance_path, vbs_performance, self.objective)

    def get_solver_data(self: SelectionOutput,
                        train_data: PerformanceDataFrame,
                        objective: SparkleObjective) -> SelectionSolverData:
        """Initalise SelectionSolverData object."""
        solver_performance_ranking = train_data.get_solver_ranking(objective=objective)
        num_solvers = train_data.num_solvers
        return SelectionSolverData(solver_performance_ranking,
                                   num_solvers)

    def serialize_solvers(self: SelectionOutput,
                          sd: SelectionSolverData) -> dict:
        """Transform SelectionSolverData to dictionary format."""
        return {
            "number_of_solvers": sd.num_solvers,
            "single_best_solver": sd.single_best_solver,
            "solver_ranking": [
                {
                    "solver_name": solver[0],
                    "performance": solver[1]
                }
                for solver in sd.solver_performance_ranking
            ]
        }

    def serialize_performance(self: SelectionOutput,
                              sp: SelectionPerformance) -> dict:
        """Transform SelectionPerformance to dictionary format."""
        return {
            "vbs_performance": sp.vbs_performance,
            "actual_performance": sp.actual_performance,
            "objective": self.objective.name,
            "metric": sp.metric
        }

    def serialize_instances(self: SelectionOutput,
                            instances: list[InstanceSet]) -> dict:
        """Transform Instances to dictionary format."""
        return {
            "number_of_instance_sets": len(instances),
            "instance_sets": [
                {
                    "name": instance.name,
                    "number_of_instances": instance.size
                }
                for instance in instances
            ]
        }

    def serialize_contribution(self: SelectionOutput) -> dict:
        """Transform marginal contribution ranking to dictionary format."""
        return {
            "marginal_contribution_actual": [
                {
                    "solver_name": ranking[0],
                    "marginal_contribution": ranking[1],
                    "best_performance": ranking[2]
                }
                for ranking in self.marginal_contribution_actual
            ],
            "marginal_contribution_perfect": [
                {
                    "solver_name": ranking[0],
                    "marginal_contribution": ranking[1],
                    "best_performance": ranking[2]
                }
                for ranking in self.marginal_contribution_perfect
            ]
        }

    def serialize_settings(self: SelectionOutput) -> dict:
        """Transform settings to dictionary format."""
        return {
            "cutoff_time": self.cutoff_time,
        }

    def write_output(self: SelectionOutput) -> None:
        """Write data into a JSON file."""
        test_data = self.serialize_instances(self.test_instances) if self.test_instances\
            else None
        output_data = {
            "solvers": self.serialize_solvers(self.solver_data),
            "training_instances": self.serialize_instances(self.training_instances),
            "test_instances": test_data,
            "performance": self.serialize_performance(self.performance_data),
            "settings": self.serialize_settings(),
            "marginal_contribution": self.serialize_contribution()
        }

        self.output.parent.mkdir(parents=True, exist_ok=True)
        with self.output.open("w") as f:
            json.dump(output_data, f, indent=4)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Configurator class to use different algorithm configurators like SMAC."""

from __future__ import annotations
from abc import abstractmethod
from pathlib import Path

from runrunner import Runner, Run
from sparkle.solver import Solver
from sparkle.solver.validator import Validator
from sparkle.instance import InstanceSet
from sparkle.types import SparkleObjective


class Configurator:
    """Abstact class to use different configurators like SMAC."""
    configurator_cli_path = Path(__file__).parent.resolve() / "configurator_cli.py"

    def __init__(self: Configurator, validator: Validator, output_path: Path,
                 executable_path: Path, configurator_target: Path,
                 objectives: list[SparkleObjective], base_dir: Path, tmp_path: Path,
                 multi_objective_support: bool = False) -> None:
        """Initialize Configurator.

        Args:
            validator: Validator object to validate configurations runs
            output_path: Output directory of the Configurator.
            executable_path: Executable of the configurator for Sparkle to call
            configurator_target: The wrapper algorithm to standardize configurator
                input/output towards solver wrappers.
            objectives: The list of Sparkle Objectives the configurator has to
                optimize.
            base_dir: Where to execute the configuration
            tmp_path: Path for the temporary files of the configurator, optional
            multi_objective_support: Whether the configurator supports
                multi objective optimization for solvers.
        """
        self.validator = validator
        self.output_path = output_path
        self.executable_path = executable_path
        self.configurator_target = configurator_target
        self.objectives = objectives
        self.base_dir = base_dir
        self.tmp_path = tmp_path
        self.multiobjective = multi_objective_support
        self.scenario = None
        if len(self.objectives) > 1 and not self.multiobjective:
            print("Warning: Multiple objectives specified but current configurator "
                  f"{self.configurator_path.name} only supports single objective. "
                  f"Defaulted to first specified objective: {self.objectives[0].name}")

    @property
    def scenario_class(self: Configurator) -> ConfigurationScenario:
        """Return the scenario class of the configurator."""
        return ConfigurationScenario

    @abstractmethod
    def configure(self: Configurator,
                  scenario: ConfigurationScenario,
                  validate_after: bool = True,
                  sbatch_options: list[str] = [],
                  num_parallel_jobs: int = None,
                  base_dir: Path = None,
                  run_on: Runner = Runner.SLURM) -> Run:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario to execute.
            validate_after: Whether to validate the configuration on the training set
                afterwards or not.
            sbatch_options: List of slurm batch options to use
            num_parallel_jobs: The maximum number of jobs to run in parallel
            base_dir: The base_dir of RunRunner where the sbatch scripts will be placed
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        raise NotImplementedError

    @abstractmethod
    def get_optimal_configuration(self: Configurator,
                                  solver: Solver,
                                  instance_set: InstanceSet,
                                  objective: SparkleObjective) -> tuple[float, str]:
        """Returns the optimal configuration string for a solver of an instance set."""
        raise NotImplementedError

    @staticmethod
    def organise_output(output_source: Path, output_target: Path) -> None | str:
        """Method to restructure and clean up after a single configurator call."""
        raise NotImplementedError

    def set_scenario_dirs(self: Configurator,
                          solver: Solver, instance_set: InstanceSet) -> None:
        """Patching method to allow the rebuilding of configuration scenario."""
        raise NotImplementedError

    def get_status_from_logs(self: Configurator) -> None:
        """Method to scan the log files of the configurator for warnings."""
        raise NotImplementedError


class ConfigurationScenario:
    """Template class to handle a configuration scenarios."""
    def __init__(self: ConfigurationScenario, solver: Solver,
                 instance_set: InstanceSet,
                 sparkle_objectives: list[SparkleObjective] = None)\
            -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
        """
        self.solver = solver
        self.instance_set = instance_set
        self.sparkle_objectives = sparkle_objectives
        self.name = f"{self.solver.name}_{self.instance_set.name}"

    def create_scenario(self: ConfigurationScenario, parent_directory: Path) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        raise NotImplementedError

    def create_scenario_file(self: ConfigurationScenario) -> None:
        """Create a file with the configuration scenario.

        Writes supplementary information to the target algorithm (algo =) as:
        algo = {configurator_target} {solver_directory} {sparkle_objective}
        """
        raise NotImplementedError

    @staticmethod
    def from_file(scenario_file: Path, solver: Solver, instance_set: InstanceSet,
                  ) -> ConfigurationScenario:
        """Reads scenario file and initalises ConfigurationScenario."""
        raise NotImplementedError



import numpy as np
import pandas as pd

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace import Configuration

from autofolio.aslib_scenario import ASlibScenario


import xgboost as xgb

__author__ = "Marius Lindauer"
__license__ = "BSD"


class XGBoost(object):

    @staticmethod
    def add_params(cs: ConfigurationSpace):
        '''
            adds parameters to ConfigurationSpace 
        '''
        
        try:
            classifier = cs.get_hyperparameter("classifier")
            if "XGBoost" not in classifier.choices:
                return 

            num_round = UniformIntegerHyperparameter(
                name="xgb:num_round", lower=10, upper=100, default_value=50, log=True)
            cs.add_hyperparameter(num_round)
            alpha = UniformFloatHyperparameter(
                name="xgb:alpha", lower=0, upper=10, default_value=1)
            cs.add_hyperparameter(alpha)
            lambda_ = UniformFloatHyperparameter(
                name="xgb:lambda", lower=1, upper=10, default_value=1)
            cs.add_hyperparameter(lambda_)
            colsample_bylevel = UniformFloatHyperparameter(
                name="xgb:colsample_bylevel", lower=0.5, upper=1, default_value=1)
            cs.add_hyperparameter(colsample_bylevel)
            colsample_bytree = UniformFloatHyperparameter(
                name="xgb:colsample_bytree", lower=0.5, upper=1, default_value=1)
            cs.add_hyperparameter(colsample_bytree)
            subsample = UniformFloatHyperparameter(
                name="xgb:subsample", lower=0.01, upper=1, default_value=1)
            cs.add_hyperparameter(subsample)
            max_delta_step = UniformFloatHyperparameter(
                name="xgb:max_delta_step", lower=0, upper=10, default_value=0)
            cs.add_hyperparameter(max_delta_step)
            min_child_weight = UniformFloatHyperparameter(
                name="xgb:min_child_weight", lower=0, upper=20, default_value=1)
            cs.add_hyperparameter(min_child_weight)
            max_depth = UniformIntegerHyperparameter(
                name="xgb:max_depth", lower=1, upper=10, default_value=6)
            cs.add_hyperparameter(max_depth)
            gamma = UniformFloatHyperparameter(
                name="xgb:gamma", lower=0, upper=10, default_value=0)
            cs.add_hyperparameter(gamma)
            eta = UniformFloatHyperparameter(
                name="xgb:eta", lower=0, upper=1, default_value=0.3)
            cs.add_hyperparameter(eta)

            cond = InCondition(
                child=num_round, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=alpha, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=lambda_, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=colsample_bylevel, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=colsample_bytree, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=subsample, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=max_delta_step, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=min_child_weight, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=max_depth, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=gamma, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
            cond = InCondition(
                child=eta, parent=classifier, values=["XGBoost"])
            cs.add_condition(cond)
        except: 
            return
        

    def __init__(self):
        '''
            Constructor
        '''

        self.model = None
        self.attr = []

    def __str__(self):
        return "XGBoost"

    def fit(self, X, y, config: Configuration, weights=None):
        '''
            fit pca object to ASlib scenario data

            Arguments
            ---------
            X: numpy.array
                feature matrix
            y: numpy.array
                label vector
            weights: numpy.array
                vector with sample weights
            config: ConfigSpace.Configuration
                configuration

        '''
        
        xgb_config = {'nthread': 1,
         'verbosity': 0, 
         'objective': 'binary:logistic',
         'seed': 12345}
        print(config, flush=True)
        for param in config:
            if param.startswith("xgb:") and config[param] is not None:
                self.attr.append("%s=%s"%(param[4:],config[param]))
            else:
                continue
            if param == "xgb:num_round":
                continue
            xgb_config[param[4:]] = config[param]
        print(xgb_config, flush=True)
        dtrain = xgb.DMatrix(X, label=y, weight=weights)

        self.model = xgb.train(xgb_config, dtrain, config["xgb:num_round"])        

    def predict(self, X):
        '''
            transform ASLib scenario data

            Arguments
            ---------
            X: numpy.array
                instance feature matrix

            Returns
            -------

        '''
        preds = np.array(self.model.predict(xgb.DMatrix(X)))
        preds[preds < 0.5] = 0
        preds[preds >= 0.5] = 1
        return preds

    def get_attributes(self):
        '''
            returns a list of tuples of (attribute,value) 
            for all learned attributes

            Returns
            -------
            list of tuples of (attribute,value) 
        '''
        return self.attr



#!/usr/bin/env python3
"""Sparkle command to remove a solver from the Sparkle platform."""

import sys
import argparse
import shutil

from sparkle.platform import file_help as sfh
from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame
from sparkle.CLI.help import logging as sl
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.SolverRemoveArgument.names,
                        **ac.SolverRemoveArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    solver_path = resolve_object_name(args.solver,
                                      gv.solver_nickname_mapping,
                                      gv.settings().DEFAULT_solver_dir)

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.REMOVE_SOLVER])
    if solver_path is None:
        print(f'Could not resolve Solver path/name "{solver_path}"!')
        sys.exit(-1)
    if not solver_path.parent == gv.settings().DEFAULT_solver_dir:
        # Allow user to only specify solvers in Sparkle solver dir
        print("Specified Solver is not is platform directory! Exiting.")
        sys.exit(-1)

    print(f"Start removing solver {solver_path.name} ...")

    solver_nickname_mapping = gv.solver_nickname_mapping
    if len(solver_nickname_mapping):
        nickname = None
        for key in solver_nickname_mapping:
            if solver_nickname_mapping[key] == str(solver_path):
                nickname = key
                break
        sfh.add_remove_platform_item(
            nickname,
            gv.solver_nickname_list_path,
            gv.file_storage_data_mapping[gv.solver_nickname_list_path],
            remove=True)

    if gv.settings().DEFAULT_performance_data_path.exists():
        performance_data = PerformanceDataFrame(
            gv.settings().DEFAULT_performance_data_path)
        if solver_path.name in performance_data.dataframe.columns:
            performance_data.remove_solver(solver_path.name)
        performance_data.save_csv()

    shutil.rmtree(solver_path)

    print(f"Removing solver {solver_path.name} done!")



"""General tools for Sparkle."""
import time
import random
import os


def get_time_pid_random_string() -> str:
    """Return a combination of time, Process ID, and random int as string.

    Returns:
      A random string composed of time, PID and a random positive integer value.
    """
    time_stamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime(time.time()))
    return f"{time_stamp}_{os.getpid()}_{int(random.getrandbits(32))}"



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions to log which output was created by Sparkle where."""
from __future__ import annotations

import time
from pathlib import Path
from pathlib import PurePath

from runrunner.logger import Log as RunRunnerLog

from sparkle.CLI.help import global_variables as gv


# Keep track of which command called Sparkle
global caller
caller: str = "unknown"

# Current caller file path
global caller_log_path
caller_log_path: str | PurePath = "not set"

# Root output directory for the calling command in the form of
# Output/<timestamp>_<command_name>/
global caller_out_dir
caller_out_dir: Path = Path(".")

# Log directory for the calling command in the form of
# Output/<timestamp>_<command_name>/Log/
global caller_log_dir
caller_log_dir: Path = Path(".")


def _update_caller(argv: list[str]) -> None:
    """Update which command is currently active.

    Args:
        argv: List containing the command line arguments derived from sys.argv.

    """
    global caller
    caller = Path(argv[0]).stem


def _update_caller_file_path(timestamp: str) -> None:
    """Create a new file path for the caller with the given timestamp.

    Args:
        timestamp: String representation of the time.

    """
    caller_file = caller + "_main_log.txt"
    caller_dir = Path(timestamp + "_" + caller)
    log_dir = gv.settings().DEFAULT_log_output
    # Set caller directory for other Sparkle functions to use
    global caller_out_dir
    caller_out_dir = Path(caller_dir)
    global caller_log_path
    caller_log_path = PurePath(log_dir / caller_out_dir / caller_file)
    global caller_log_dir
    caller_log_dir = log_dir / caller_out_dir

    # Create needed directories if they don't exist
    caller_dir = Path(caller_log_path).parents[0]
    caller_dir.mkdir(parents=True, exist_ok=True)
    caller_log_dir.mkdir(parents=True, exist_ok=True)

    # If the caller output file does not exist yet, write the header
    if not Path(caller_log_path).is_file():
        output_header = "\t Timestamp\t\t\t\t\t\t\t  Path\t\t\t\t\t\t\t Description\n"
        with Path(caller_log_path).open("a") as output_file:
            output_file.write(output_header)


def add_output(output_path: str, description: str) -> None:
    """Add output location and description to the log of the current command.

    Args:
        output_path: The file path of where output is written to.
        description: A short description of what kind of output is written to this file.

    """
    # Prepare logging information
    timestamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime(time.time()))
    output_str = f"{timestamp}\t{output_path}\t{description}\n"
    # Write output path and description to caller file
    with Path(caller_log_path).open("a") as output_file:
        output_file.write(output_str)


def log_command(argv: list[str]) -> None:
    """Write to file which command was executed.

    Includes information on when it was executed, with which arguments, and
    where details about it's output are stored (if any).

    Args:
        argv: List containing the command line arguments derived from sys.argv.

    """
    # Determine caller
    _update_caller(argv)

    # Prepare logging information
    timestamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime(time.time()))
    _update_caller_file_path(timestamp)
    output_file = caller_log_path
    args = " ".join(argv[0:])
    log_str = timestamp + "   " + args + "   " + str(output_file) + "\n"

    # If the log file does not exist yet, write the header
    log_path = gv.settings().DEFAULT_output / "sparkle.log"
    if not log_path.is_file():
        log_header = "\t Timestamp\t\t\t\t\t\t\t  Command\t\t\t\t\t\t\t Output details\n"
        log_str = log_header + log_str

    # Write to log file
    log_path.open("a").write(log_str)

    # Pipe RunRunner log to the caller log
    RunRunnerLog.set_log_file(caller_log_path)



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions to record and restore a Sparkle platform."""
import shutil
import sys
import os
import time
from pathlib import Path
import zipfile

from sparkle.CLI.help import global_variables as gv
from sparkle.tools.general import get_time_pid_random_string


def save_current_sparkle_platform() -> None:
    """Store the current Sparkle platform in a .zip file."""
    time_stamp = time.strftime("%Y-%m-%d-%H:%M:%S", time.localtime(time.time()))
    snapshot_tmp_path = gv.settings().DEFAULT_snapshot_dir /\
        f"Snapshot_{os.getlogin()}_{time_stamp}"
    snapshot_tmp_path.mkdir(parents=True)  # Create temporary directory for zip
    for working_dir in gv.settings().DEFAULT_working_dirs:
        if working_dir.exists():
            shutil.copytree(working_dir, snapshot_tmp_path / working_dir.name)
    shutil.make_archive(snapshot_tmp_path, "zip", snapshot_tmp_path)
    shutil.rmtree(snapshot_tmp_path)
    print(f"Snapshot file {snapshot_tmp_path}.zip saved successfully!")


def remove_current_platform() -> None:
    """Remove the current Sparkle platform."""
    for working_dir in gv.settings().DEFAULT_working_dirs:
        shutil.rmtree(working_dir, ignore_errors=True)


def create_working_dirs() -> None:
    """Create working directories."""
    for working_dir in gv.settings().DEFAULT_working_dirs:
        working_dir.mkdir(parents=True, exist_ok=True)


def extract_snapshot(snapshot_file: Path) -> None:
    """Restore a Sparkle platform from a snapshot.

    Args:
      snapshot_file: Path to the where the current Sparkle platform should be stored.
    """
    tmp_directory = Path(f"tmp_directory_{get_time_pid_random_string()}")
    gv.settings().DEFAULT_tmp_output.mkdir(exist_ok=True)
    with zipfile.ZipFile(snapshot_file, "r") as zip_ref:
        zip_ref.extractall(tmp_directory)
    shutil.copytree(tmp_directory, "./", dirs_exist_ok=True)
    shutil.rmtree(tmp_directory)


def load_snapshot(snapshot_file: Path) -> None:
    """Load a Sparkle platform from a snapshot.

    Args:
        snapshot_file: File path to the file where the Sparkle platform is stored.
    """
    if not snapshot_file.exists():
        print(f"ERROR: Snapshot file {snapshot_file} does not exist!")
        sys.exit(-1)
    if not snapshot_file.suffix == ".zip":
        print(f"ERROR: File {snapshot_file} is not a .zip file!")
        sys.exit(-1)
    print("Cleaning existing Sparkle platform ...")
    remove_current_platform()
    print("Existing Sparkle platform cleaned!")

    print(f"Loading snapshot file {snapshot_file} ...")
    extract_snapshot(snapshot_file)
    print(f"Snapshot file {snapshot_file} loaded successfully!")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
import sys
import time
from pathlib import Path

from runrunner import Runner

from sparkle.solver import Solver
from sparkle.types import resolve_objective


if __name__ == "__main__":
    # Incoming call from SMAC:
    # Translate input to Solver object input
    argsiter = iter(sys.argv[8:])
    args = zip(argsiter, argsiter)
    configuration = {arg.strip("-"): val for arg, val in args}
    # Args 1-7 conditions of the run, the rest are configurations for the solver
    # [Solver_dir, SparkleObjective, instance, specifics, cutoff_time, runlength, seed]
    solver_dir = Path(sys.argv[1])
    objective = resolve_objective(sys.argv[2])
    instance = sys.argv[3]
    cutoff_time = float(sys.argv[5])
    seed = int(sys.argv[7])

    runsolver_binary = solver_dir / "runsolver"
    solver = Solver(solver_dir,
                    raw_output_directory=Path(),
                    runsolver_exec=runsolver_binary)
    # Call Runsolver with the solver configurator wrapper and its arguments
    start_t = time.time()
    output = solver.run(instance=instance,
                        objectives=[objective],
                        seed=seed,
                        cutoff_time=cutoff_time,
                        configuration=configuration,
                        run_on=Runner.LOCAL)
    run_time = min(time.time() - start_t, cutoff_time)

    # Return values to SMAC
    # SMAC2 does not accept nan values etc for quality
    quality = "0"
    if not objective.time and objective.name in output.keys():
        quality = float(output[objective.name])
        if not objective.minimise:
            quality = -1 * quality
    print("Result for SMAC: "
          f"{output['status']}, {output['cpu_time']}, 0, {quality}, {seed}")



import numpy as np
import pandas as pd

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace import Configuration

from autofolio.aslib_scenario import ASlibScenario


import sklearn.ensemble

__author__ = "Marius Lindauer"
__license__ = "BSD"


class RandomForestRegressor(object):

    @staticmethod
    def add_params(cs: ConfigurationSpace):
        '''
            adds parameters to ConfigurationSpace 
        '''

        try:
            regressor = cs.get_hyperparameter("regressor")
            if "RandomForestRegressor" not in regressor.choices:
                return

            n_estimators = UniformIntegerHyperparameter(
                name="rfreg:n_estimators", lower=10, upper=100, default_value=10, log=True)
            cs.add_hyperparameter(n_estimators)
            max_features = CategoricalHyperparameter(
                name="rfreg:max_features", choices=["sqrt", "log2", "None"], default_value="sqrt")
            cs.add_hyperparameter(max_features)
            max_depth = UniformIntegerHyperparameter(
                name="rfreg:max_depth", lower=10, upper=2 ** 31, default_value=2 ** 31, log=True)
            cs.add_hyperparameter(max_depth)
            min_samples_split = UniformIntegerHyperparameter(
                name="rfreg:min_samples_split", lower=2, upper=100, default_value=2, log=True)
            cs.add_hyperparameter(min_samples_split)
            min_samples_leaf = UniformIntegerHyperparameter(
                name="rfreg:min_samples_leaf", lower=2, upper=100, default_value=10, log=True)
            cs.add_hyperparameter(min_samples_leaf)
            bootstrap = CategoricalHyperparameter(
                name="rfreg:bootstrap", choices=[True, False], default_value=True)
            cs.add_hyperparameter(bootstrap)

            cond = InCondition(
                child=n_estimators, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)
            cond = InCondition(
                child=max_features, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)
            cond = InCondition(
                child=max_depth, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)
            cond = InCondition(
                child=min_samples_split, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)
            cond = InCondition(
                child=min_samples_leaf, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)
            cond = InCondition(
                child=bootstrap, parent=regressor, values=["RandomForestRegressor"])
            cs.add_condition(cond)

        except:
            return

    def __init__(self):
        '''
            Constructor
        '''

        self.model = None

    def __str__(self):
        return "RandomForestRegressor"

    def fit(self, X, y, config: Configuration):
        '''
            fit pca object to ASlib scenario data

            Arguments
            ---------
            X: numpy.array
                feature matrix
            y: numpy.array
                label vector
            weights: numpy.array
                vector with sample weights
            config: ConfigSpace.Configuration
                configuration

        '''

        self.model = sklearn.ensemble.RandomForestRegressor(n_estimators=config["rfreg:n_estimators"],
                                            max_features=config[
                                                "rfreg:max_features"] if config[
                                                "rfreg:max_features"] != "None" else None,
                                            max_depth=config["rfreg:max_depth"],
                                            min_samples_split=config[
                                                "rfreg:min_samples_split"],
                                            min_samples_leaf=config[
                                                "rfreg:min_samples_leaf"],
                                            bootstrap=config["rfreg:bootstrap"],
                                            random_state=12345)
        self.model.fit(X, y)

    def predict(self, X):
        '''
            transform ASLib scenario data

            Arguments
            ---------
            X: numpy.array
                instance feature matrix

            Returns
            -------

        '''

        return self.model.predict(X)
    
    def get_attributes(self):
        '''
            returns a list of tuples of (attribute,value) 
            for all learned attributes
            
            Returns
            -------
            list of tuples of (attribute,value) 
        '''
        attr = []
        attr.append("max_depth = %d" % (self.model.max_depth))
        attr.append("min_samples_split = %d" % (self.model.min_samples_split))
        attr.append("min_samples_leaf = %d" % (self.model.min_samples_leaf))
        attr.append("n_estimators = %d" % (self.model.n_estimators))
        attr.append("max_features = %s" % (self.model.max_features))
        return attr
        



import logging
import traceback

import numpy as np
import pandas as pd

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace import Configuration

from autofolio.aslib_scenario import ASlibScenario

__author__ = "Marius Lindauer"
__license__ = "BSD"


class JointRegression(object):

    @staticmethod
    def add_params(cs: ConfigurationSpace):
        '''
            adds parameters to ConfigurationSpace 
        '''

        selector = cs.get_hyperparameter("selector")
        regressor = cs.get_hyperparameter("regressor")
        if "JointRegressor" in selector.choices:
            cond = InCondition(child=regressor, parent=selector, values=["JointRegressor"])
            cs.add_condition(cond)

    def __init__(self, regressor_class):
        '''
            Constructor
        '''
        self.regressors = []
        self.logger = logging.getLogger("JointRegressor")
        self.regressor_class = regressor_class

    def fit(self, scenario: ASlibScenario, config: Configuration):
        '''
            fit pca object to ASlib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas
            config: ConfigSpace.Configuration
                configuration
        '''
        self.logger.info("Fit PairwiseRegressor with %s" %
                         (self.regressor_class))

        self.algorithms = scenario.algorithms

        n_algos = len(scenario.algorithms)
        X = scenario.feature_data.values
        Xs = None
        Ys = None
        
        for i in range(n_algos):
            y = scenario.performance_data[scenario.algorithms[i]].values
            x_i = np.zeros((X.shape[0], n_algos))
            x_i[:,i] = 1
            X_i = np.hstack([X,x_i])
            if i == 0:
                Xs = X_i
                Ys = y
            else:
                Xs = np.vstack([Xs,X_i])
                Ys = np.hstack([Ys,y])
            
        reg = self.regressor_class()
        reg.fit(Xs, Ys, config)
        self.regressors = reg

    def predict(self, scenario: ASlibScenario):
        '''
            predict schedules for all instances in ASLib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas

            Returns
            -------
                schedule: {inst -> (solver, time)}
                    schedule of solvers with a running time budget
        '''

        if scenario.algorithm_cutoff_time:
            cutoff = scenario.algorithm_cutoff_time
        else:
            cutoff = 2**31

        n_algos = len(scenario.algorithms)
        X = scenario.feature_data.values
        scores = np.zeros((X.shape[0], n_algos))
        for i in range(n_algos):
            x_i = np.zeros((X.shape[0], n_algos))
            x_i[:,i] = 1
            X_i = np.hstack([X,x_i])
            Y = self.regressors.predict(X_i)
            scores[:, i] += Y

        #self.logger.debug(
        #   sorted(list(zip(scenario.algorithms, scores)), key=lambda x: x[1], reverse=True))
        algo_indx = np.argmin(scores, axis=1)
        
        schedules = dict((str(inst),[s]) for s,inst in zip([(scenario.algorithms[i], cutoff+1) for i in algo_indx], scenario.feature_data.index))
        #self.logger.debug(schedules)
        return schedules

    def get_attributes(self):
        '''
            returns a list of tuples of (attribute,value) 
            for all learned attributes
            
            Returns
            -------
            list of tuples of (attribute,value) 
        '''
        reg_attr = self.regressors[0].get_attributes()
        attr = [{self.regressor_class.__name__:reg_attr}]

        return attr


#!/usr/bin/env python3
"""Sparkle command to execute ablation analysis."""

import argparse
import sys
from pathlib import PurePath

from runrunner.base import Runner

from sparkle.solver.ablation import AblationScenario
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.solver import Solver
from sparkle.instance import instance_set
from sparkle.CLI.help import argparse_custom as ac
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description=("Runs parameter importance between the default and configured "
                     "parameters with ablation. This command requires a finished "
                     "configuration for the solver instance pair."),
        epilog=("Note that if no test instance set is given, the validation is performed"
                " on the training set."))
    parser.add_argument("--solver", required=False, type=str, help="path to solver")
    parser.add_argument(*ac.InstanceSetTrainAblationArgument.names,
                        **ac.InstanceSetTrainAblationArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTestAblationArgument.names,
                        **ac.InstanceSetTestAblationArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.TargetCutOffTimeAblationArgument.names,
                        **ac.TargetCutOffTimeAblationArgument.kwargs)
    parser.add_argument(*ac.WallClockTimeArgument.names,
                        **ac.WallClockTimeArgument.kwargs)
    parser.add_argument(*ac.NumberOfRunsAblationArgument.names,
                        **ac.NumberOfRunsAblationArgument.kwargs)
    parser.add_argument(*ac.RacingArgument.names,
                        **ac.RacingArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    parser.set_defaults(ablation_settings_help=False)
    return parser


if __name__ == "__main__":
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    solver_path = resolve_object_name(args.solver,
                                      gv.solver_nickname_mapping,
                                      gv.settings().DEFAULT_solver_dir)
    solver = Solver(solver_path)
    instance_set_train = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)
    instance_set_test = resolve_object_name(
        args.instance_set_test,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.RUN_ABLATION])

    if ac.set_by_user(args, "settings_file"):
        # Do first, so other command line options can override settings from the file
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "objectives"):
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "target_cutoff_time"):
        gv.settings().set_general_target_cutoff_time(
            args.target_cutoff_time, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "wallclock_time"):
        gv.settings().set_config_wallclock_time(
            args.wallclock_time, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "number_of_runs"):
        gv.settings().set_config_number_of_runs(
            args.number_of_runs, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "racing"):
        gv.settings().set_ablation_racing_flag(
            args.number_of_runs, SettingState.CMD_LINE
        )

    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    instance_set_train_name = instance_set_train.name
    configurator = gv.settings().get_general_sparkle_configurator()
    configurator.set_scenario_dirs(solver, instance_set_train)
    if instance_set_test is not None:
        instance_set_test_name = instance_set_test.name
    else:
        instance_set_test = instance_set_train
        instance_set_test_name = instance_set_train.name

    if not configurator.scenario.result_directory.is_dir():
        print("Error: No configuration results found for the given solver and training"
              " instance set. Ablation needs to have a target configuration.")
        print("Please run configuration first")
        sys.exit(-1)
    else:
        print("Configuration exists!")

    ablation_scenario = AblationScenario(
        solver, instance_set_train, instance_set_test,
        gv.settings().DEFAULT_ablation_output,
        gv.settings().DEFAULT_ablation_exec,
        gv.settings().DEFAULT_ablation_validation_exec, override_dirs=True)

    # Instances
    ablation_scenario.create_instance_file()
    ablation_scenario.create_instance_file(test=True)

    # Configurations
    ablation_scenario.create_configuration_file()
    print("Submiting ablation run...")
    runs = ablation_scenario.submit_ablation(run_on=run_on)

    if run_on == Runner.LOCAL:
        print("Ablation analysis finished!")
    else:
        job_id_str = ",".join([run.run_id for run in runs])
        print(f"Ablation analysis running. Waiting for Slurm job(s) with id(s): "
              f"{job_id_str}")



#!/usr/bin/env python3
"""Sparkle command to run solvers to get their performance data."""
from __future__ import annotations

import sys
import argparse
from pathlib import PurePath, Path

import runrunner as rrr
from runrunner.base import Runner, Run

from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument(*ac.RecomputeRunSolversArgument.names,
                        **ac.RecomputeRunSolversArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.TargetCutOffTimeRunSolversArgument.names,
                        **ac.TargetCutOffTimeRunSolversArgument.kwargs)
    parser.add_argument(*ac.AlsoConstructSelectorAndReportArgument.names,
                        **ac.AlsoConstructSelectorAndReportArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)

    return parser


def running_solvers_performance_data(
        performance_data_csv_path: Path,
        num_job_in_parallel: int,
        rerun: bool = False,
        run_on: Runner = Runner.SLURM) -> Run:
    """Run the solvers for the performance data.

    Parameters
    ----------
    performance_data_csv_path: Path
        The path to the performance data file
    num_job_in_parallel: int
        The maximum number of jobs to run in parallel
    rerun: bool
        Run only solvers for which no data is available yet (False) or (re)run all
        solvers to get (new) performance data for them (True)
    run_on: Runner
        Where to execute the solvers. For available values see runrunner.base.Runner
        enum. Default: "Runner.SLURM".

    Returns
    -------
    run: runrunner.LocalRun or runrunner.SlurmRun
        If the run is local return a QueuedRun object with the information concerning
        the run.
    """
    # Open the performance data csv file
    performance_dataframe = PerformanceDataFrame(performance_data_csv_path)
    # List of jobs to do
    jobs = performance_dataframe.get_job_list(rerun=rerun)
    num_jobs = len(jobs)

    cutoff_time_str = str(gv.settings().get_general_target_cutoff_time())

    print(f"Cutoff time for each solver run: {cutoff_time_str} seconds")
    print(f"Total number of jobs to run: {num_jobs}")

    # If there are no jobs, stop
    if num_jobs == 0:
        return None

    if run_on == Runner.LOCAL:
        print("Running the solvers locally")
    elif run_on == Runner.SLURM:
        print("Running the solvers through Slurm")

    sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
    srun_options = ["-N1", "-n1"] + sbatch_options
    objectives = gv.settings().get_general_sparkle_objectives()
    run_solvers_core = Path(__file__).parent.resolve() / "core" / "run_solvers_core.py"
    cmd_list = [f"{run_solvers_core} "
                f"--performance-data {performance_data_csv_path} "
                f"--instance {inst_p} --solver {solver_p} "
                f"--objectives {','.join([str(o) for o in objectives])} "
                f"--log-dir {sl.caller_log_dir}" for inst_p, _, solver_p in jobs]

    run = rrr.add_to_queue(
        runner=run_on,
        cmd=cmd_list,
        parallel_jobs=num_job_in_parallel,
        name=CommandName.RUN_SOLVERS,
        base_dir=sl.caller_log_dir,
        sbatch_options=sbatch_options,
        srun_options=srun_options)

    if run_on == Runner.LOCAL:
        # TODO: It would be nice to extract some info per job and print it
        # As the user now only sees jobs starting and completing without their results
        run.wait()

    return run


def run_solvers_on_instances(
        recompute: bool = False,
        run_on: Runner = Runner.SLURM,
        also_construct_selector_and_report: bool = False) -> None:
    """Run all the solvers on all the instances that were not not previously run.

    If recompute is True, rerun everything even if previously run. Where the solvers are
    executed can be controlled with "run_on".

    Parameters
    ----------
    recompute: bool
        If True, recompute all solver-instance pairs even if they were run before.
        Default: False
    run_on: Runner
        On which computer or cluster environment to run the solvers.
        Available: Runner.LOCAL, Runner.SLURM. Default: Runner.SLURM
    also_construct_selector_and_report: bool
        If True, the selector will be constructed and a report will be produced.
    """
    if recompute:
        PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path).clean_csv()
    num_job_in_parallel = gv.settings().get_number_of_jobs_in_parallel()

    runs = [running_solvers_performance_data(
        performance_data_csv_path=gv.settings().DEFAULT_performance_data_path,
        num_job_in_parallel=num_job_in_parallel,
        rerun=recompute,
        run_on=run_on)]

    # If there are no jobs return
    if all(run is None for run in runs):
        print("Running solvers done!")
        return

    sbatch_user_options = gv.settings().get_slurm_extra_options(as_args=True)
    if also_construct_selector_and_report:
        runs.append(rrr.add_to_queue(
            runner=run_on,
            cmd="sparkle/CLI/construct_portfolio_selector.py",
            name=CommandName.CONSTRUCT_PORTFOLIO_SELECTOR,
            dependencies=runs[-1],
            base_dir=sl.caller_log_dir,
            sbatch_options=sbatch_user_options))

        runs.append(rrr.add_to_queue(
            runner=run_on,
            cmd="sparkle/CLI/generate_report.py",
            name=CommandName.GENERATE_REPORT,
            dependencies=runs[-1],
            base_dir=sl.caller_log_dir,
            sbatch_options=sbatch_user_options))

    if run_on == Runner.LOCAL:
        print("Waiting for the local calculations to finish.")
        for run in runs:
            if run is not None:
                run.wait()
        print("Running solvers done!")
    elif run_on == Runner.SLURM:
        print("Running solvers. Waiting for Slurm job(s) with id(s): "
              f'{",".join(r.run_id for r in runs if r is not None)}')


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    if args.settings_file is not None:
        # Do first, so other command line options can override settings from the file
        gv.settings().read_settings_ini(args.settings_file, SettingState.CMD_LINE)

    if args.objectives is not None:
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE
        )

    if args.target_cutoff_time is not None:
        gv.settings().set_general_target_cutoff_time(
            args.target_cutoff_time, SettingState.CMD_LINE)

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.RUN_SOLVERS])

    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    print("Start running solvers ...")

    # Write settings to file before starting, since they are used in callback scripts
    gv.settings().write_used_settings()

    run_solvers_on_instances(
        recompute=args.recompute,
        also_construct_selector_and_report=args.also_construct_selector_and_report,
        run_on=run_on)



"""Test functionalities related to the about command."""

from __future__ import annotations
from unittest import TestCase
from sparkle import about


class TestAbout(TestCase):
    """Tests function of sparkle.about."""
    def test_about_str_type(self: TestCase) -> None:
        """Tests about_str() returns the correct type."""
        # Simple test, mostly for example purpose
        self.assertIsInstance(about.about_str, str)
        self.assertIsInstance(about.name, str)
        self.assertIsInstance(about.version, str)



"""Class for Sparkle Objective and Performance."""
from __future__ import annotations
from enum import Enum
import typing
import numpy as np


class UseTime(str, Enum):
    """Use time or not."""
    WALL_TIME = "WALL_TIME"
    CPU_TIME = "CPU_TIME"
    NO = "NO"

    @classmethod
    def _missing_(cls: UseTime, value: object) -> UseTime:
        """Return error use time."""
        return UseTime.NO


class SparkleObjective:
    """Objective for Sparkle specified by user."""

    name: str
    run_aggregator: typing.Callable
    instance_aggregator: typing.Callable
    solver_aggregator: typing.Callable
    minimise: bool
    post_process: typing.Callable
    use_time: UseTime

    def __init__(self: SparkleObjective,
                 name: str,
                 run_aggregator: typing.Callable = np.mean,
                 instance_aggregator: typing.Callable = np.mean,
                 solver_aggregator: typing.Callable = None,
                 minimise: bool = True,
                 post_process: typing.Callable = None,
                 use_time: UseTime = UseTime.NO) -> None:
        """Create sparkle objective from string."""
        self.name = name
        self.run_aggregator: typing.Callable = run_aggregator
        self.instance_aggregator: typing.Callable = instance_aggregator
        if solver_aggregator is None:
            solver_aggregator = np.min if minimise else np.max
        self.solver_aggregator: typing.Callable = solver_aggregator
        self.minimise: bool = minimise
        self.post_process: typing.Callable = post_process
        self.use_time: UseTime = use_time

    def __str__(self: SparkleObjective) -> str:
        """Return a stringified version."""
        return f"{self.name}"

    @property
    def time(self: SparkleObjective) -> bool:
        """Return whether the objective is time based."""
        return self.use_time != UseTime.NO


class PAR(SparkleObjective):
    """Penalised Averaged Runtime Objective for Sparkle."""

    def __init__(self: PAR, k: int = 10) -> None:
        """Initialize PAR."""
        self.k = k
        if k <= 0:
            raise ValueError("k must be greater than 0.")

        def penalise(value: float, cutoff: float) -> float:
            """Return penalised value."""
            if value > cutoff:
                return cutoff * self.k
            return value

        super().__init__(f"PAR{k}", use_time=UseTime.CPU_TIME, post_process=penalise)



#!/usr/bin/env python3
"""Sparkle command to add a solver to the Sparkle platform."""
import os
import stat
import sys
import argparse
import shutil
from pathlib import Path

import runrunner as rrr

from sparkle.platform import file_help as sfh
from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame
from sparkle.CLI.run_solvers import running_solvers_performance_data
from sparkle.solver import Solver
from sparkle.CLI.help import logging as sl
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.platform.settings_objects import SettingState


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Add a solver to the Sparkle platform.",
        epilog="")
    parser.add_argument(*ac.DeterministicArgument.names,
                        **ac.DeterministicArgument.kwargs)
    parser.add_argument(*ac.RunSolverNowArgument.names,
                        **ac.RunSolverNowArgument.kwargs)
    parser.add_argument(*ac.NicknameSolverArgument.names,
                        **ac.NicknameSolverArgument.kwargs)
    parser.add_argument(*ac.SolverPathArgument.names,
                        **ac.SolverPathArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    parser.add_argument(*ac.SkipChecksArgument.names,
                        **ac.SkipChecksArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    solver_source = Path(args.solver_path)
    deterministic = args.deterministic

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.ADD_SOLVER])

    if not solver_source.exists():
        print(f'Solver path "{solver_source}" does not exist!')
        sys.exit(-1)

    nickname = args.nickname

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    if args.run_checks:
        print("Running checks...")
        solver = Solver(Path(solver_source))
        pcs_file = solver.get_pcs_file()
        if pcs_file is None:
            print("None or multiple .pcs files found. Solver "
                  "is not valid for configuration.")
        else:
            print(f"One pcs file detected: {pcs_file.name}. ", end="")
            if solver.read_pcs_file():
                print("Can read the pcs file.")
            else:
                print("WARNING: Can not read the provided pcs file format.")

        configurator_wrapper_path = solver_source / Solver.wrapper
        if not (configurator_wrapper_path.is_file()
                and os.access(configurator_wrapper_path, os.X_OK)):
            print(f"WARNING: Solver {solver_source.name} does not have a solver wrapper "
                  f"(Missing file {Solver.wrapper}) or is not executable. ")

    # Start add solver
    solver_directory = gv.settings().DEFAULT_solver_dir / solver_source.name
    if not solver_directory.exists():
        solver_directory.mkdir(parents=True, exist_ok=True)
    else:
        print(f"ERROR: Solver {solver_source.name} already exists! "
              "Can not add new solver.")
        sys.exit(-1)
    shutil.copytree(solver_source, solver_directory, dirs_exist_ok=True)
    # Save the deterministic bool in the solver
    with (solver_directory / Solver.meta_data).open("w+") as fout:
        fout.write(str({"deterministic": deterministic}))

    # Add RunSolver executable to the solver
    runsolver_path = gv.settings().DEFAULT_runsolver_exec
    if runsolver_path.name in [file.name for file in solver_directory.iterdir()]:
        print("Warning! RunSolver executable detected in Solver "
              f"{solver_source.name}. This will be replaced with "
              f"Sparkle's version of RunSolver. ({runsolver_path})")
    runsolver_target = solver_directory / runsolver_path.name
    shutil.copyfile(runsolver_path, runsolver_target)
    runsolver_target.chmod(os.stat(runsolver_target).st_mode | stat.S_IEXEC)

    performance_data = PerformanceDataFrame(
        gv.settings().DEFAULT_performance_data_path,
        objectives=gv.settings().get_general_sparkle_objectives())
    performance_data.add_solver(solver_directory)
    performance_data.save_csv()

    print(f"Adding solver {solver_source.name} done!")

    if nickname is not None:
        sfh.add_remove_platform_item(solver_directory,
                                     gv.solver_nickname_list_path, key=nickname)

    if args.run_solver_now:
        num_job_in_parallel = gv.settings().get_number_of_jobs_in_parallel()
        dependency_run_list = [running_solvers_performance_data(
            gv.settings().DEFAULT_performance_data_path, num_job_in_parallel,
            rerun=False, run_on=run_on
        )]

        sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
        srun_options = ["-N1", "-n1"] + sbatch_options
        run_construct_portfolio_selector = rrr.add_to_queue(
            cmd="sparkle/CLI/construct_portfolio_selector.py",
            name=CommandName.CONSTRUCT_PORTFOLIO_SELECTOR,
            dependencies=dependency_run_list,
            base_dir=sl.caller_log_dir,
            sbatch_options=sbatch_options,
            srun_options=srun_options)

        dependency_run_list.append(run_construct_portfolio_selector)

        run_generate_report = rrr.add_to_queue(
            cmd="sparkle/CLI/generate_report.py",
            name=CommandName.GENERATE_REPORT,
            dependencies=dependency_run_list,
            base_dir=sl.caller_log_dir,
            sbatch_options=sbatch_options,
            srun_options=srun_options)

    # Write used settings to file
    gv.settings().write_used_settings()



"""Test public methods of configuration scenario class."""

from __future__ import annotations

import shutil

from unittest import TestCase
from unittest.mock import patch
from unittest.mock import Mock
from pathlib import Path

from sparkle.configurator.implementations import SMAC2, SMAC2Scenario
from sparkle.solver import Solver
from sparkle.instance import instance_set
from sparkle.types.objective import PAR


class TestConfigurationScenario(TestCase):
    """Class bundling all tests regarding ConfigurationScenario."""
    def setUp(self: TestConfigurationScenario) -> None:
        """Setup executed before each test."""
        self.solver_path = Path("tests", "test_files", "Solvers", "Test-Solver")
        self.solver = Solver(self.solver_path)

        self.instance_set = instance_set(
            Path("tests/test_files/Instances/Test-Instance-Set"))
        self.run_number = 2

        self.parent_directory = Path("tests/test_files/test_configurator")
        self.parent_directory.mkdir(parents=True, exist_ok=True)

        self.instance_file_directory = Path("tests/test_files/test_configurator"
                                            "/scenarios/instances/Test-Instance-Set")
        self.instance_file_directory.mkdir(parents=True, exist_ok=True)
        self.wallclock_time = 600
        self.cutoff_time = 60
        self.cutoff_length = "max"
        self.sparkle_objective = PAR(10)
        self.configurator = SMAC2([self.sparkle_objective], Path(), Path())
        self.scenario = SMAC2Scenario(
            solver=self.solver,
            instance_set=self.instance_set,
            number_of_runs=self.run_number,
            wallclock_time=self.wallclock_time,
            cutoff_time=self.cutoff_time,
            cutoff_length=self.cutoff_length,
            sparkle_objectives=[self.sparkle_objective],
            use_features=False,
            configurator_target=Path(self.configurator.target_algorithm))

    def tearDown(self: TestConfigurationScenario) -> None:
        """Cleanup executed after each test."""
        shutil.rmtree(self.parent_directory, ignore_errors=True)

    def test_configuration_scenario_init(self: TestConfigurationScenario) -> None:
        """Test if all variables that are set in the init are correct."""
        self.assertEqual(self.scenario.solver, self.solver)
        self.assertEqual(self.scenario.instance_set.directory,
                         self.instance_set.directory)
        self.assertFalse(self.scenario.use_features)
        self.assertEqual(self.scenario.feature_data, None)
        self.assertEqual(self.scenario.name,
                         f"{self.solver.name}_{self.instance_set.name}")

    def test_configuration_scenario_check_scenario_directory(
        self: TestConfigurationScenario
    ) -> None:
        """Test if create_scenario() correctly creates the scenario directory."""
        self.scenario.create_scenario(self.parent_directory)

        self.assertTrue(self.scenario.directory.is_dir())
        self.assertEqual((self.scenario.directory
                          / "outdir_train_configuration").is_dir(),
                         True)
        self.assertTrue((self.scenario.directory / "tmp").is_dir())

    def test_configuration_scenario_check_result_directory(
        self: TestConfigurationScenario,
    ) -> None:
        """Test if create_scenario() creates the result directory."""
        self.scenario.create_scenario(self.parent_directory)

        self.assertTrue(self.scenario.result_directory.is_dir())

    def test_configuration_scenario_check_instance_directory(
        self: TestConfigurationScenario
    ) -> None:
        """Test if create_scenario() creates the instance directory."""
        self.scenario.create_scenario(self.parent_directory)
        self.assertTrue(self.scenario.instance_set.directory.is_dir())

    @patch("pathlib.Path.absolute")
    def test_configuration_scenario_check_scenario_file(
        self: TestConfigurationScenario,
        mock_abs_path: Mock
    ) -> None:
        """Test if create_scenario() correctly creates the scenario file."""
        inst_list_path = Path("tests/test_files/test_configurator/scenarios/instances/"
                              "Test-Instance-Set/Test-Instance-Set_train.txt")
        mock_abs_path.side_effect = [Path("tests/test_files/test_configurator"),
                                     Path("/configurator_dir/target_algorithm.py"),
                                     self.solver_path,
                                     inst_list_path,
                                     inst_list_path,
                                     Path(),
                                     Path()]
        self.scenario.create_scenario(self.parent_directory)

        reference_scenario_file = Path("tests", "test_files", "reference_files",
                                       "scenario_file.txt")

        # Use to show full diff of file
        self.maxDiff = None
        self.assertTrue(self.scenario.scenario_file_path.is_file())
        output = self.scenario.scenario_file_path.open().read()
        # Strip the output of the homedirs (Due to absolute paths)
        output = output.replace(str(Path.home()), "")
        self.assertEqual(output,
                         reference_scenario_file.open().read())



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for parallel portfolio report generation."""
from pathlib import Path
import csv
import operator

import plotly.express as px
import pandas as pd
import plotly.io as pio

from sparkle.platform import latex as stex
from sparkle.platform import generate_report_for_selection as sgfs
from sparkle.types import SparkleObjective, SolverStatus
from sparkle.instance import InstanceSet


pio.kaleido.scope.mathjax = None  # Bug fix for kaleido


def get_solver_list_latex(solver_list: list[str]) -> str:
    """Return the list of solvers as string, including each solver-seed combination.

    Args:
        solver_list: The solver list to convert

    Returns:
        A list of solvers in the parallel portfolio as str.
    """
    latex_itemize = ""

    for solver_str in solver_list:
        # Solver string may contain path and variation seed
        solver_split = solver_str.split(" ")
        solver_name = Path(solver_split[0]).name
        solver_seeds = int(solver_split[2]) if len(solver_split) == 3 else 0

        latex_itemize += f"\\item \\textbf{{{stex.underscore_for_latex(solver_name)}}}\n"
        # Only include if we used more than one seed
        if solver_seeds > 1:
            seeds = ",".join(list[range(1, solver_seeds + 1)])
            latex_itemize += f"\\item[]With seeds: {seeds}\n"

    return latex_itemize


def get_portfolio_metrics(
        solver_list: list[str],
        instance_set: InstanceSet,
        results: dict[list[str, str]],
        objective: SparkleObjective) -> tuple[dict[str, float], str, dict[str, float]]:
    """Return the portfolio metrics for SBS, aggregated results per solver and VBS.

    Args:
        solver_list: list of solvers
        instance_list: List of paths to instance sets
        results: dictionary of results of the portfolio

    Returns:
        A quatro-tuple:
            A dict containing the objective value per instance for the single best solver
            A string containing the name of the single best solver.
            A second dict with aggrated objective values over all instances per solver.
            A third dict containing the virtual best solver (Portfolio) per instance.
    """
    corrected_solver_results = {solver: [] for solver in solver_list}
    instance_worst = None
    op = operator.gt if objective.minimise else operator.lt
    for instance in results:
        for (solver, status, value) in results[instance]:
            if value == "None":
                corrected_solver_results[solver].append(None)
                continue  # Solver failed to return a value
            value = float(value)
            if instance_worst is None or op(value, instance_worst):
                instance_worst = value
            if SolverStatus(status) != SolverStatus.SUCCESS:
                # If the solver wasn't successful, we also assign the worst known value
                corrected_solver_results[solver].append(None)
                continue
            corrected_solver_results[solver].append(float(value))
        for solver in corrected_solver_results:
            if corrected_solver_results[solver][-1] is None:
                corrected_solver_results[solver][-1] = instance_worst
    aggregated_results_solvers = {solver: objective.instance_aggregator(
        corrected_solver_results[solver]) for solver in corrected_solver_results}
    vbs_results_instance = {instance: objective.solver_aggregator(
        [corrected_solver_results[solver][index] for solver in corrected_solver_results])
        for index, instance in enumerate(instance_set._instance_names)}
    # Find the single best solver (SBS)
    sbs_name = min(aggregated_results_solvers, key=aggregated_results_solvers.get)
    sbs_results = corrected_solver_results[sbs_name]
    sbs_dict = {name: sbs_results[index]
                for index, name in enumerate(instance_set._instance_names)}
    return sbs_dict, sbs_name, aggregated_results_solvers, vbs_results_instance


def get_figure_parallel_portfolio_sparkle_vs_sbs(
        target_directory: Path,
        solver_list: list[str],
        instance_set: InstanceSet,
        results: list[str],
        objective: SparkleObjective) -> tuple[
        str, dict[str, float], dict[str, float]]:
    """Generate PaP vs SBS figure and return a string to include it in LaTeX.

    Args:
        target_directory: Path where to place the files
        parallel_portfolio_path: Parallel portfolio path.
        instances: List of instances.

    Returns:
        a three tuple:
            str_value: A string to include the PaP vs SBS figure in LaTeX
            dict_all_solvers: A dict containing the penalised average run time per
                solver.
            dict_actual_parallel_portfolio_penalty_time_on_each_instance: A dict with
                instance names and the penalised running time of the PaP.
    """
    sbs_instance_results, sbs_solver, dict_all_solvers, vbs_instance_results =\
        get_portfolio_metrics(solver_list, instance_set, results, objective)
    figure_filename = "figure_parallel_portfolio_sparkle_vs_sbs"
    data = [[sbs_instance_results[instance], vbs_instance_results[instance]]
            for instance in sbs_instance_results]
    generate_figure(target_directory,
                    f"SBS ({stex.underscore_for_latex(sbs_solver)})",
                    "Parallel-Portfolio", figure_filename, objective.name, data)
    latex_include = f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"
    return latex_include, dict_all_solvers, vbs_instance_results


def get_results_table(results: dict[str, str, str],
                      dict_all_solvers: dict[str, float], parallel_portfolio_path: Path,
                      dict_portfolio: dict[str, float],
                      solver_with_solutions: dict[str, int],
                      n_unsolved_instances: int, n_instances: int,
                      performance_metric: str) -> str:
    """Returns a LaTeX table with the portfolio results.

    Args:
        results: The total results with status and runtime per solver
        dict_all_solvers: A dict containing the penalised average run time per solver.
        parallel_portfolio_path: Parallel portfolio path
        dict_portfolio: A dict with instance names and the penalised running time of the
            PaP.
        solver_with_solutions: A dict with solver name as key, and number of solved
            instances for the corresponding solver as value, see
            get_solvers_with_solution.
        n_unsolved_instances: Number of unsolved instances.
        n_instances: Number of instances.

    Returns:
        A string containing LaTeX code for a table with the portfolio results.
    """
    portfolio_par = 0.0
    for instance in dict_portfolio:
        portfolio_par += dict_portfolio[instance]
    portfolio_par = portfolio_par / n_instances
    total_killed = 0
    for instance in results:
        for (_, status, _) in results[instance]:
            total_killed += (status.lower() == "killed")
    # Table 1: Portfolio results
    table_string = (
        "\\caption{\\textbf{Portfolio results}} \\label{tab:portfolio_results} ")
    table_string += "\\begin{tabular}{rrrrr}"
    table_string += (
        "\\textbf{Portfolio nickname} & \\textbf{"
        f"{performance_metric}"
        "} & \\textbf{\\#Timeouts} & "
        "\\textbf{\\#Cancelled} & \\textbf{\\# Solved} \\\\ \\hline ")
    table_string += (
        f"{stex.underscore_for_latex(parallel_portfolio_path.name)} & "
        f"{round(portfolio_par,2)} & {n_unsolved_instances} & {total_killed} & "
        f"{n_instances-n_unsolved_instances} \\\\ ")
    table_string += "\\end{tabular}"
    table_string += "\\bigskip"
    # Table 2: Solver results
    table_string += "\\caption{\\textbf{Solver results}} \\label{tab:solver_results} "
    table_string += "\\begin{tabular}{rrrrr}"

    for i, line in enumerate(dict_all_solvers):
        solver_name = Path(line).name

        if i == 0:
            table_string += (
                "\\textbf{Solver} & \\textbf{"
                f"{performance_metric}"
                "} & \\textbf{\\#Timeouts} & "
                "\\textbf{\\#Cancelled} & \\textbf{\\#Best solver} \\\\ \\hline ")

        if solver_name not in solver_with_solutions:
            cancelled = n_instances - n_unsolved_instances
            table_string += (
                f"{stex.underscore_for_latex(solver_name)} & "
                f"{round(dict_all_solvers[line], 2)} & {n_unsolved_instances} & "
                f"{cancelled} & 0 \\\\ ")
        else:
            cancelled = (n_instances - n_unsolved_instances
                         - solver_with_solutions[solver_name])
            table_string += (
                f"{stex.underscore_for_latex(solver_name)} & "
                f"{round(dict_all_solvers[line], 2)} & {n_unsolved_instances} & "
                f"{cancelled} & {solver_with_solutions[solver_name]} \\\\ ")
    table_string += "\\end{tabular}"
    return table_string


def parallel_report_variables(target_directory: Path,
                              parallel_portfolio_path: Path,
                              bibliograpghy_path: Path,
                              objective: SparkleObjective,
                              cutoff: int,
                              instance_set: InstanceSet) -> dict[str, str]:
    """Returns a mapping between LaTeX report variables and their values.

    Args:
        target_directory: Path to where to place the generated files.
        parallel_portfolio_path: Parallel portfolio path.
        bibliograpghy_path: Path to the bib file
        instances: List of instances.

    Returns:
        A dictionary that maps variables used in the LaTeX report to values.
    """
    variables_dict = {"bibliographypath": bibliograpghy_path.absolute(),
                      "cutoffTime": cutoff,
                      "performanceMetric": objective.name,
                      "numInstanceClasses": "1"}  # Currently no support for multi sets
    # Get the results data
    csv_data = [line for line in
                csv.reader((parallel_portfolio_path / "results.csv").open("r"))]
    header = csv_data[0]
    csv_data = csv_data[1:]
    solver_column = header.index("Solver")
    instance_column = header.index("Instance")
    status_column = header.index("status")
    objective_column = header.index(objective.name)
    solver_list = list(set([line[solver_column]
                            for line in csv_data]))  # Unique set of solvers
    results = {name: [] for name in instance_set._instance_names}
    for row in csv_data:
        if row[instance_column] in results.keys():
            results[row[instance_column]].append(
                [row[solver_column], row[status_column], row[objective_column]])

    variables_dict["numSolvers"] = len(solver_list)
    variables_dict["solverList"] = get_solver_list_latex(solver_list)
    variables_dict["instanceClassList"] =\
        sgfs.get_instance_set_count_list(instance_set._instance_paths)

    # Produce some statistics on the parallel portfolio
    solvers_solutions = {solver: 0 for solver in solver_list}
    instance_names_copy = instance_set._instance_names.copy()
    for line in csv_data:
        if (line[instance_column] in instance_names_copy
                and SolverStatus(line[status_column]) == SolverStatus.SUCCESS):
            solvers_solutions[line[solver_column]] += 1
            instance_names_copy.remove(line[instance_column])
    unsolved_instances = instance_set.size - sum([solvers_solutions[key]
                                                  for key in solvers_solutions])
    inst_succes = []
    for solver in solvers_solutions:
        inst_succes.append("\\item Solver "
                           f"\\textbf{{{stex.underscore_for_latex(solver)}}}, was the "
                           "best solver on "
                           f"\\textbf{{{solvers_solutions[solver]}}} instance(s)")
    if unsolved_instances > 0:
        inst_succes.append(f"\\item \\textbf{{{unsolved_instances}}} instances(s) "
                           "remained unsolved")

    variables_dict["solversWithSolution"] = "\n".join(inst_succes)

    (figure_name, dict_all_solvers,
        dict_actual_parallel_portfolio_penalty_time_on_each_instance) =\
        get_figure_parallel_portfolio_sparkle_vs_sbs(target_directory, solver_list,
                                                     instance_set, results, objective)

    variables_dict["figure-parallel-portfolio-sparkle-vs-sbs"] = figure_name
    variables_dict["resultsTable"] = get_results_table(
        results, dict_all_solvers, parallel_portfolio_path,
        dict_actual_parallel_portfolio_penalty_time_on_each_instance,
        solvers_solutions, unsolved_instances, instance_set.size, objective.name)

    if objective.time:
        variables_dict["decisionBool"] = "\\decisiontrue"
    else:
        variables_dict["decisionBool"] = "\\decisionfalse"
    return variables_dict


def generate_figure(
        target_directory: Path,
        sbs_name: str, parallel_portfolio_name: str,
        figure_parallel_portfolio_vs_sbs_filename: str,
        performance_measure: str, data: list) -> None:
    """Generates image for parallel portfolio report."""
    upper_bound = max([x for xs in data for x in xs]) * 1.5
    lower_bound = 0.01

    output_plot = target_directory / f"{figure_parallel_portfolio_vs_sbs_filename}.pdf"

    xlabel = f"{sbs_name}, {performance_measure}"
    ylabel = f"{parallel_portfolio_name}"
    df = pd.DataFrame(data, columns=[xlabel, ylabel])
    fig = px.scatter(data_frame=df, x=xlabel, y=ylabel,
                     range_x=[lower_bound, upper_bound],
                     range_y=[lower_bound, upper_bound],
                     log_x=True, log_y=True,
                     width=500, height=500)
    # Add in the seperation line
    fig.add_shape(type="line", x0=0, y0=0, x1=upper_bound, y1=upper_bound,
                  line=dict(color="lightgrey", width=1))
    fig.update_traces(marker=dict(color="RoyalBlue", symbol="x"))
    fig.update_layout(
        plot_bgcolor="white"
    )
    fig.update_xaxes(
        type="log",
        mirror=True,
        dtick=1,
        ticks="outside",
        showline=True,
        showgrid=True,
        linecolor="black",
        gridcolor="lightgrey"
    )
    fig.update_yaxes(
        type="log",
        mirror=True,
        dtick=1,
        ticks="outside",
        showline=True,
        showgrid=True,
        linecolor="black",
        gridcolor="lightgrey"
    )
    fig.write_image(output_plot)


def generate_report_parallel_portfolio(parallel_portfolio_path: Path,
                                       target_path: Path,
                                       latex_template: Path,
                                       bibliograpghy_path: Path,
                                       objective: SparkleObjective,
                                       cutoff: int,
                                       instances: InstanceSet) -> None:
    """Generate a report for a parallel algorithm portfolio.

    Args:
        parallel_portfolio_path: Parallel portfolio path.
        target_path: Where the report data will be placed.
        latex_template: Path to the latex template path used
        bibliograpghy_path: Path to the bib file
        objective: The objective of the portfolio
        cutoff: The cutoff time for each solver
        instances: List of instances.
    """
    target_path.mkdir(parents=True, exist_ok=True)
    dict_variable_to_value = parallel_report_variables(
        target_path, parallel_portfolio_path, bibliograpghy_path, objective,
        cutoff, instances)

    stex.generate_report(latex_template,
                         "template-Sparkle-for-parallel-portfolio.tex",
                         target_path,
                         "Sparkle_Report_Parallel_Portfolio",
                         dict_variable_to_value)



#!/usr/bin/env python3
"""Sparkle command to generate a report for an executed experiment."""

import sys
import argparse
from pathlib import Path, PurePath

from sparkle.CLI.help import global_variables as gv
from sparkle.solver.ablation import AblationScenario
from sparkle.platform import generate_report_for_selection as sgfs
from sparkle.platform import \
    generate_report_for_configuration as sgrfch
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.platform import \
    generate_report_for_parallel_portfolio as sgrfpph
from sparkle.solver import Solver
from sparkle.solver.validator import Validator
from sparkle.instance import instance_set
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.platform.output.configuration_output import ConfigurationOutput
from sparkle.platform.output.selection_output import SelectionOutput
from sparkle.platform.output.parallel_portfolio_output import ParallelPortfolioOutput

from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description=("Without any arguments a report for the most recent algorithm "
                     "selection or algorithm configuration procedure is generated."),
        epilog=("Note that if a test instance set is given, the training instance set "
                "must also be given."))
    # Configuration arguments
    parser.add_argument(*ac.SolverReportArgument.names,
                        **ac.SolverReportArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTrainReportArgument.names,
                        **ac.InstanceSetTrainReportArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTestReportArgument.names,
                        **ac.InstanceSetTestReportArgument.kwargs)
    parser.add_argument(*ac.NoAblationReportArgument.names,
                        **ac.NoAblationReportArgument.kwargs)
    # Selection arguments
    parser.add_argument(*ac.SelectionReportArgument.names,
                        **ac.SelectionReportArgument.kwargs)
    parser.add_argument(*ac.TestCaseDirectoryArgument.names,
                        **ac.TestCaseDirectoryArgument.kwargs)
    # Common arguments
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*ac.GenerateJSONArgument.names,
                        **ac.GenerateJSONArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))

    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    selection = args.selection
    test_case_dir = args.test_case_directory
    only_json = args.only_json

    solver = resolve_object_name(args.solver,
                                 gv.solver_nickname_mapping,
                                 gv.settings().DEFAULT_solver_dir, Solver)
    instance_set_train = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)
    instance_set_test = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.GENERATE_REPORT])

    # Do first, so other command line options can override settings from the file
    if ac.set_by_user(args, "settings_file"):
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )
    if args.objectives is not None:
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE)

    Settings.check_settings_changes(gv.settings(), prev_settings)
    # If no arguments are set get the latest scenario
    if not selection and test_case_dir is None and solver is None:
        scenario = gv.latest_scenario().get_latest_scenario()
        if scenario == Scenario.SELECTION:
            selection = True
            test_case_dir = gv.latest_scenario().get_selection_test_case_directory()
        elif scenario == Scenario.CONFIGURATION:
            solver = gv.latest_scenario().get_config_solver()
            instance_set_train = gv.latest_scenario().get_config_instance_set_train()
            instance_set_test = gv.latest_scenario().get_config_instance_set_test()
        elif scenario == Scenario.PARALLEL_PORTFOLIO:
            parallel_portfolio_path = gv.latest_scenario().get_parallel_portfolio_path()
            pap_instance_set =\
                gv.latest_scenario().get_parallel_portfolio_instance_set()

    flag_instance_set_train = instance_set_train is not None
    flag_instance_set_test = instance_set_test is not None

    # Reporting for algorithm selection
    if selection or test_case_dir is not None:
        objective = gv.settings().get_general_sparkle_objectives()[0]
        if not objective.time:
            print("ERROR: The selection report is not implemented for "
                  " non-runtime objectives!")
            sys.exit(-1)
        selection_scenario = gv.latest_scenario().get_selection_scenario_path()
        actual_portfolio_selector_path = selection_scenario / "portfolio_selector"
        if not actual_portfolio_selector_path.is_file():
            print("Before generating a Sparkle report, please first construct the "
                  "Sparkle portfolio selector. Not generating a Sparkle report, stopping"
                  " execution!")
            sys.exit(-1)

        print("Generating report for selection...")
        train_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
        feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
        test_data = None
        test_case_path = Path(test_case_dir) if test_case_dir is not None else None
        if test_case_dir is not None and (test_case_path
                                          / "performance_data.csv").exists():
            test_data = PerformanceDataFrame(test_case_path / "performance_data.csv")
        # Create machine readable selection output
        instance_folders = set(Path(instance).parent
                               for instance in train_data.instances)
        instance_sets = []
        for dir in instance_folders:
            set = instance_set(dir)
            instance_sets.append(set)
        test_set = None if test_case_dir is None else instance_set(Path(test_case_dir))
        cutoff_time = gv.settings().get_general_target_cutoff_time()
        output = gv.settings().DEFAULT_selection_output_analysis
        selection_output = SelectionOutput(
            selection_scenario, train_data, feature_data,
            instance_sets, test_set, objective, cutoff_time,
            output)
        selection_output.write_output()
        print("Machine readable output is placed at: ", selection_output.output)

        if not only_json:
            sgfs.generate_report_selection(
                gv.settings().DEFAULT_selection_output_analysis,
                gv.settings().DEFAULT_latex_source,
                "template-Sparkle-for-selection.tex",
                gv.settings().DEFAULT_latex_bib,
                gv.settings().DEFAULT_extractor_dir,
                selection_scenario,
                feature_data,
                train_data,
                objective,
                gv.settings().get_general_extractor_cutoff_time(),
                gv.settings().get_general_target_cutoff_time(),
                test_data
            )
            if test_case_dir is None:
                print("Report generated ...")
            else:
                print("Report for test generated ...")

    elif gv.latest_scenario().get_latest_scenario() == Scenario.PARALLEL_PORTFOLIO:
        # Reporting for parallel portfolio
        # Machine readable Output
        cutoff_time = gv.settings().get_general_target_cutoff_time()
        objective = gv.settings().get_general_sparkle_objectives()[0]
        output = gv.settings().DEFAULT_parallel_portfolio_output_analysis
        parallel_portfolio_output = ParallelPortfolioOutput(parallel_portfolio_path,
                                                            pap_instance_set,
                                                            objective,
                                                            output)
        parallel_portfolio_output.write_output()
        print("Machine readable output is placed at: ", parallel_portfolio_output.output)

        if not only_json:
            sgrfpph.generate_report_parallel_portfolio(
                parallel_portfolio_path,
                gv.settings().DEFAULT_parallel_portfolio_output_analysis,
                gv.settings().DEFAULT_latex_source,
                gv.settings().DEFAULT_latex_bib,
                gv.settings().get_general_sparkle_objectives()[0],
                gv.settings().get_general_target_cutoff_time(),
                pap_instance_set)
            print("Parallel portfolio report generated ...")
    else:
        # Reporting for algorithm configuration
        if solver is None:
            print("Error! No Solver found for configuration report generation.")
            sys.exit(-1)

        # If only the testing set is given return an error
        if not flag_instance_set_train and flag_instance_set_test:
            print("Argument Error! Only a testing set was provided, please also "
                  "provide a training set")
            print(f"Usage: {sys.argv[0]} --solver <solver> [--instance-set-train "
                  "<instance-set-train>] [--instance-set-test <instance-set-test>]")
            sys.exit(-1)
        instance_set_train_name = instance_set_train.name
        gv.settings().get_general_sparkle_configurator()\
            .set_scenario_dirs(solver, instance_set_train)
        # Generate a report depending on which instance sets are provided
        if flag_instance_set_train or flag_instance_set_test:
            # Check if there are result to generate a report from
            validator = Validator(gv.settings().DEFAULT_validation_output)
            train_res = validator.get_validation_results(
                solver, instance_set_train)
            if instance_set_test is not None:
                test_res = validator.get_validation_results(solver,
                                                            instance_set_test)
            if len(train_res) == 0 or (instance_set_test is not None
                                       and len(test_res) == 0):
                print("Error: Results not found for the given solver and instance set(s)"
                      ' combination. Make sure the "configure_solver" and "validate_'
                      'configured_vs_default" commands were correctly executed. ')
                sys.exit(-1)
        else:
            print("Error: No results from validate_configured_vs_default found that "
                  "can be used in the report!")
            sys.exit(-1)
        # Extract config scenario data for report, but this should be read from the
        # scenario file instead as we can't know wether features were used or not now
        number_of_runs = gv.settings().get_config_number_of_runs()
        solver_calls = gv.settings().get_config_solver_calls()
        cpu_time = gv.settings().get_config_cpu_time()
        wallclock_time = gv.settings().get_config_wallclock_time()
        cutoff_time = gv.settings().get_general_target_cutoff_time()
        cutoff_length = gv.settings().get_configurator_target_cutoff_length()
        sparkle_objectives =\
            gv.settings().get_general_sparkle_objectives()
        configurator = gv.settings().get_general_sparkle_configurator()
        configurator.scenario = configurator.scenario_class(
            solver, instance_set_train, number_of_runs, solver_calls, cpu_time,
            wallclock_time, cutoff_time, cutoff_length, sparkle_objectives)
        configurator.scenario._set_paths(configurator.output_path)
        ablation_scenario = None
        if args.flag_ablation:
            ablation_scenario = AblationScenario(
                solver, instance_set_train, instance_set_test,
                gv.settings().DEFAULT_ablation_output)

        # Create machine readable output
        solver_name = gv.latest_scenario().get_config_solver().name
        instance_set_name = gv.latest_scenario().get_config_instance_set_train().name
        output = gv.settings().DEFAULT_configuration_output_analysis
        config_output = ConfigurationOutput(configurator.scenario.directory,
                                            solver, configurator,
                                            instance_set_train,
                                            instance_set_test,
                                            output)
        config_output.write_output()
        print("Machine readable output is placed at: ", config_output.output)

        if not only_json:
            sgrfch.generate_report_for_configuration(
                solver,
                gv.settings().get_general_sparkle_configurator(),
                Validator(gv.settings().DEFAULT_validation_output),
                gv.settings().DEFAULT_extractor_dir,
                gv.settings().DEFAULT_configuration_output_analysis,
                gv.settings().DEFAULT_latex_source,
                gv.settings().DEFAULT_latex_bib,
                instance_set_train,
                gv.settings().get_general_extractor_cutoff_time(),
                instance_set_test,
                ablation=ablation_scenario
            )

    # Write used settings to file
    gv.settings().write_used_settings()



"""File containing the Validator class."""

from __future__ import annotations

import sys
from pathlib import Path
import csv
import ast
import runrunner as rrr
from runrunner import Runner, Run

from sparkle.solver import Solver
from sparkle.instance import InstanceSet
from sparkle.types import SparkleObjective, resolve_objective
from sparkle.tools.runsolver_parsing import get_solver_args


class Validator():
    """Class to handle the validation of solvers on instance sets."""
    def __init__(self: Validator,
                 out_dir: Path = Path(),
                 tmp_out_dir: Path = Path()) -> None:
        """Construct the validator."""
        self.out_dir = out_dir
        self.tmp_out_dir = tmp_out_dir

    def validate(self: Validator,
                 solvers: list[Path] | list[Solver] | Solver | Path,
                 configurations: list[dict] | dict | Path,
                 instance_sets: list[InstanceSet],
                 objectives: list[SparkleObjective],
                 cut_off: int,
                 subdir: Path = None,
                 dependency: list[Run] | Run = None,
                 sbatch_options: list[str] = [],
                 run_on: Runner = Runner.SLURM) -> Run:
        """Validate a list of solvers (with configurations) on a set of instances.

        Args:
            solvers: list of solvers to validate
            configurations: list of configurations for each solver we validate.
                If a path is supplied, will use each line as a configuration.
            instance_sets: set of instance sets on which we want to validate each solver
            objectives: list of objectives to validate
            cut_off: maximum run time for the solver per instance
            subdir: The subdir where to place the output in the outputdir. If None,
                a semi-unique combination of solver_instanceset is created.
            dependency: Jobs to wait for before executing the validation.
            sbatch_options: list of slurm batch options
            run_on: whether to run on SLURM or local
        """
        if not isinstance(solvers, list) and isinstance(configurations, list):
            # If we receive one solver but multiple configurations, we cas the
            # Solvers to a list of the same length
            solvers = [solvers] * len(configurations)
        elif not isinstance(configurations, list) and isinstance(solvers, list):
            # If there is only one configuration, we cast it to a list of the same
            # length as the solver list
            configurations = [configurations] * len(solvers)
        if not isinstance(solvers, list) or len(configurations) != len(solvers):
            print("Error: Number of solvers and configurations does not match!")
            sys.exit(-1)
        # Ensure we have the object representation of solvers
        solvers = [Solver(s) if isinstance(s, Path) else s for s in solvers]
        cmds = []
        out_paths = []
        for index, (solver, config) in enumerate(zip(solvers, configurations)):
            if config is None:
                config = {}
            elif isinstance(config, Path):
                # Point to the config line in file
                config = {"config_path": config}
            for instance_set in instance_sets:
                if subdir is None:
                    out_path = self.out_dir / f"{solver.name}_{instance_set.name}"
                else:
                    out_path = self.out_dir / subdir
                out_path.mkdir(exist_ok=True)
                for instance_path in instance_set._instance_paths:
                    cmds.append(" ".join(
                        solver.build_cmd(instance=instance_path.absolute(),
                                         objectives=objectives,
                                         seed=index,
                                         cutoff_time=cut_off,
                                         configuration=config)))
                out_paths.extend([out_path] * len(instance_set._instance_paths))
        return rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            name="validation",
            base_dir=self.tmp_out_dir,
            path=out_paths,
            dependencies=dependency,
            sbatch_options=sbatch_options,
        )

    def retrieve_raw_results(self: Validator,
                             solver: Solver,
                             instance_sets: InstanceSet | list[InstanceSet],
                             subdir: Path = None,
                             log_dir: Path = None) -> None:
        """Checks the raw results of a given solver for a specific instance_set.

        Writes the raw results to a unified CSV file for the resolve/instance_set
        combination.

        Args:
            solver: The solver for which to check the raw result path
            instance_sets: The set of instances for which to retrieve the results
            subdir: Subdir where the CSV is to be placed, passed to the append method.
            log_dir: The directory to search for log files. If none, defaults to
                the log directory of the Solver.
        """
        if isinstance(instance_sets, InstanceSet):
            instance_sets = [instance_sets]
        if log_dir is None:
            log_dir = solver.raw_output_directory
        for res in log_dir.iterdir():
            if res.suffix != ".rawres":
                continue
            solver_args = get_solver_args(res.with_suffix(".log"))
            solver_args = ast.literal_eval(solver_args)
            instance_path = Path(solver_args["instance"])
            # Remove default args
            if "config_path" in solver_args:
                # The actual solver configuration can be found elsewhere
                row_idx = int(solver_args["seed"])
                config_path = Path(solver_args["config_path"])
                if not config_path.exists():
                    config_path = log_dir / config_path
                config_str = config_path.open("r").readlines()[row_idx]
                solver_args = Solver.config_str_to_dict(config_str)
            else:
                for def_arg in ["instance", "solver_dir", "cutoff_time",
                                "seed", "objectives"]:
                    if def_arg in solver_args:
                        del solver_args[def_arg]
            solver_args = str(solver_args).replace('"', "'")

            for instance_set in instance_sets:
                if instance_path.name in instance_set._instance_names:
                    out_dict = Solver.parse_solver_output(
                        "",
                        ["-o", res.name,
                         "-v", res.with_suffix(".val").name,
                         "-w", res.with_suffix(".log").name],
                        log_dir)
                    self.append_entry_to_csv(solver.name,
                                             solver_args,
                                             instance_set,
                                             instance_path.name,
                                             solver_output=out_dict,
                                             subdir=subdir)
                    res.unlink()
                    res.with_suffix(".val").unlink(missing_ok=True)
                    res.with_suffix(".log").unlink(missing_ok=True)

    def get_validation_results(self: Validator,
                               solver: Solver,
                               instance_set: InstanceSet,
                               source_dir: Path = None,
                               subdir: Path = None,
                               config: str = None) -> list[list[str]]:
        """Query the results of the validation of solver on instance_set.

        Args:
            solver: Solver object
            instance_set: Instance set
            source_dir: Path where to look for any unprocessed output.
                By default, look in the solver's tmp dir.
            subdir: Path where to place the .csv file subdir. By default will be
                'self.outputdir/solver.name_instanceset.name/validation.csv'
            config: Path to the configuration if the solver was configured, None
                    otherwise
        Returns
            A list of row lists with string values
        """
        if source_dir is None:
            source_dir = self.out_dir / f"{solver.name}_{instance_set.name}"
        if any(x.suffix == ".rawres" for x in source_dir.iterdir()):
            self.retrieve_raw_results(
                solver, instance_set, subdir=subdir, log_dir=source_dir)
        if subdir is None:
            subdir = Path(f"{solver.name}_{instance_set.name}")
        csv_file = self.out_dir / subdir / "validation.csv"
        csv_data = [line for line in csv.reader(csv_file.open("r"))]
        header = csv_data[0]
        if config is not None:
            # We filter on the config string by subdict
            if isinstance(config, str):
                config = Solver.config_str_to_dict(config)
            csv_data = [line for line in csv_data[1:] if
                        config.items() == ast.literal_eval(line[1]).items()]
            csv_data.insert(0, header)
        return csv_data

    def append_entry_to_csv(self: Validator,
                            solver: str,
                            config_str: str,
                            instance_set: InstanceSet,
                            instance: str,
                            solver_output: dict,
                            subdir: Path = None) -> None:
        """Append a validation result as a row to a CSV file."""
        if subdir is None:
            subdir = Path(f"{solver}_{instance_set.name}")
        out_dir = self.out_dir / subdir
        if not out_dir.exists():
            out_dir.mkdir(parents=True)
        csv_file = out_dir / "validation.csv"
        status = solver_output["status"]
        cpu_time = solver_output["cpu_time"]
        wall_time = solver_output["wall_time"]
        del solver_output["status"]
        del solver_output["cpu_time"]
        del solver_output["wall_time"]
        sorted_keys = sorted(solver_output)
        objectives = [resolve_objective(key) for key in sorted_keys]
        objectives = [o for o in objectives if o is not None]
        if not csv_file.exists():
            # Write header
            header = ["Solver", "Configuration", "InstanceSet", "Instance", "Status",
                      "CPU Time", "Wallclock Time"] + [o.name for o in objectives]
            with csv_file.open("w") as out:
                csv.writer(out).writerow((header))
        values = [solver, config_str, instance_set.name, instance, status, cpu_time,
                  wall_time] + [solver_output[o.name] for o in objectives]
        with csv_file.open("a") as out:
            writer = csv.writer(out)
            writer.writerow(values)



"""This package provides Sparkle's wrappers for Pandas DataFrames."""
from sparkle.structures.feature_dataframe import FeatureDataFrame
from sparkle.structures.performance_dataframe import PerformanceDataFrame



import os
import setuptools

console_scripts = [
    'autofolio=autofolio.autofolio:main'
]

with open("autofolio/__version__.py") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

setuptools.setup(
    name="autofolio",
    version=version,
    author="Marius Lindauer",
    author_email="lindauer@cs.uni-freiburg.de",
    description=("AutoFolio 2, an automaticalliy configured algorithm selector."),
    license="2-clause BSD",
    keywords="algortithm selection",
    url="",
    packages=setuptools.find_packages(exclude=['test', 'source']),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: 2-clause BSD",
    ],
    platforms=['Linux'],
    tests_require=['mock',
                   'nose'],
    test_suite='nose.collector',
    entry_points = {
        'console_scripts': console_scripts
    }
)
